#include "link.h"
#include "gldefs.h"
#include "glmodulelist.h"

#define LINK_TOSTART  0
#define LINK_STARTED  1

#define LINK_DIRECTION_NORMAL  0
#define LINK_DIRECTION_REVERSE 1

CLinkList::CLinkList(TTheme *i_theme){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering constructor\n");
#endif
  
  NbLinks = 0;
  lStatus  = LINK_TOSTART;
  Links = NULL;
  theme = i_theme;
}


CLinkList::~CLinkList(){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering destructor\n");
#endif
  
  TLink *Tmp;

  while (Links!=NULL){
    Tmp=(TLink*) Links->Next;
    delete Links;
    Links=Tmp;
  };
}

int CLinkList::IsLinkStarted(){
  return (lStatus==LINK_STARTED ? 1 : 0);
}

// Start a new link if none already started and finish the link
// started if lStatus = LINK_STARTED
int CLinkList::Process(CglModule *Module, int Channel, int Mode){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering Process\n");
#endif

  if (Module==NULL){
    // we cancel an eventually started link.
    // Note that since memory allocation is only performed at the end of the creation
    // process, we only need to reset status and stop drawing link
    if (lStatus==LINK_STARTED){
      wxLogMessage("Link: cancelling");
      lStatus=LINK_TOSTART;
      return LINK_DONE;
    }
    lStatus=LINK_TOSTART;
    return LINK_OK;
  }

  if ((Mode==LINK_MODE_DELETE_I) || (Mode==LINK_MODE_DELETE_O)){
    DeleteModuleLinks(Module, Channel, Mode);
    return LINK_DONE;
  }

  if ((Mode==LINK_MODE_FEEDBACK_I) || (Mode==LINK_MODE_FEEDBACK_O)){
    ToggleFeedback(Module, Channel, Mode);
    return LINK_DONE;
  }

  if (lStatus==LINK_TOSTART){
    // determine the direction
    if (Mode==LINK_MODE_O)
      Direction = LINK_DIRECTION_NORMAL;
    else
      Direction = LINK_DIRECTION_REVERSE;

    // start a new link
    wxLogMessage("Link: starting a link");
    Start(Module, Channel);
    return LINK_OK;
  }

  if (lStatus==LINK_STARTED){
    // finish the started link
    if (((Mode==LINK_MODE_O) && (Direction==LINK_DIRECTION_NORMAL)) || ((Mode==LINK_MODE_I) && (Direction==LINK_DIRECTION_REVERSE))){
      wxLogMessage("Link: cannot link two inputs or two outputs. Cancelling started link");
      lStatus=LINK_TOSTART;
      return LINK_ERROR;
    }
    //wxLogMessage("Link: finishing the link");
    Finish(Module, Channel);
    return LINK_DONE;
  }

  return LINK_ERROR;
}


// Draw links
int CLinkList::Render(wxDC *dc, int UseTextures){
  TLink *dLink;
  int FromX=0, FromY=0, ToX=0, ToY=0;

  dLink=Links;
  if (UseTextures)
    dc->SetPen(wxPen(wxColour(theme->LinkColor[0],
			      theme->LinkColor[1],
			      theme->LinkColor[2]), 
		     2, wxSOLID));
  else
    dc->SetPen(wxPen(wxColour(255,0,0),2,wxSOLID));

  while (dLink!=NULL){
    // draw the currentlink
    FromX=(dLink->From->x+dLink->FromCh)*SIZE_OF_GRID-SIZE_OF_GRID/2;
    ToX  =(dLink->To->x  +dLink->ToCh  )*SIZE_OF_GRID-SIZE_OF_GRID/2;
    FromY=(dLink->From->y)*SIZE_OF_GRID+5*SIZE_OF_GRID/2;
    ToY  =(dLink->To->y  )*SIZE_OF_GRID-  SIZE_OF_GRID/2;

    if (dLink->State==LINK_STATE_FEEDBACK){
      if (UseTextures)
	dc->SetPen(wxPen(wxColour(theme->LinkColor[0],
				  theme->LinkColor[1],
				  theme->LinkColor[2]), 
			 2, wxLONG_DASH));
      else
	dc->SetPen(wxPen(wxColour(255,0,0),2,wxLONG_DASH));
    }
    dc->DrawLine(FromX, FromY, ToX, ToY);
    if (dLink->State==LINK_STATE_FEEDBACK){
      if (UseTextures)
	dc->SetPen(wxPen(wxColour(theme->LinkColor[0],
				  theme->LinkColor[1],
				  theme->LinkColor[2]), 
			 2, wxSOLID));
      else
	dc->SetPen(wxPen(wxColour(255,0,0),2,wxSOLID));
    }

    // step to the next link
    dLink=(TLink*) dLink->Next;
  };

  // draw an eventual started link
  if (lStatus == LINK_STARTED){
    FromX = (StartModule->x+StartChannel-1)*SIZE_OF_GRID+SIZE_OF_GRID/2;
    ToX   = ActiveLinkX*SIZE_OF_GRID+SIZE_OF_GRID/2;
    FromY = (StartModule->y+(Direction==LINK_DIRECTION_REVERSE ? -1 : StartModule->ny))
      *SIZE_OF_GRID+SIZE_OF_GRID/2;
    ToY   = ActiveLinkY*SIZE_OF_GRID+SIZE_OF_GRID/2;
    dc->DrawLine(FromX, FromY, ToX, ToY);
  }

  return 0;
}


// start a new link
int CLinkList::Start(CglModule *Module, int Channel){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering Start\n");
#endif
  // check if link in reverse direction does not
  // starting on an already connected channel.
  if (Direction==LINK_DIRECTION_REVERSE){
    TLink *TmpCheck;
    TmpCheck=Links;
    while (TmpCheck!=NULL){
      if ((TmpCheck->To==Module) && (TmpCheck->ToCh==Channel)){
	wxLogMessage("Link: cannot link twice to the same input");
	//lStatus=LINK_TOSTART;
	return 0;
      }
      TmpCheck=(TLink*) TmpCheck->Next;
    };
  }

  StartModule = Module;
  StartChannel = Channel;
  ActiveLinkX = Module->x+Channel-1;
  ActiveLinkY = Module->y+(Direction==LINK_DIRECTION_REVERSE ? -1 : Module->ny);

  lStatus = LINK_STARTED;

  return 0;
}


// finish the start link
int CLinkList::Finish(CglModule *Module, int Channel, int state){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering Finish\n");
#endif

  TLink *Tmp;

  // check if link in normal direction does not
  // finish on an already connected channel.
  if (Direction==LINK_DIRECTION_NORMAL){
    TLink *TmpCheck;
    TmpCheck=Links;
    while (TmpCheck!=NULL){
      if ((TmpCheck->To==Module) && (TmpCheck->ToCh==Channel)){
	wxLogMessage("Link: cannot link twice to the same input");
	lStatus=LINK_TOSTART;
	return 0;
      }
      TmpCheck=(TLink*) TmpCheck->Next;
    };
  }

  Tmp=new TLink;

  Tmp->Next = Links;
  Links = Tmp;
  Links->State = state;
  if (Direction==LINK_DIRECTION_NORMAL){
    Links->From    = StartModule;
    Links->FromCh  = StartChannel;
    Links->To      = Module;
    Links->ToCh    = Channel;
    wxLogMessage("Link: added in normal direction.");
  } else {
    Links->From    = Module;
    Links->FromCh  = Channel;
    Links->To      = StartModule;
    Links->ToCh    = StartChannel;
    wxLogMessage("Link: added in reverse direction.");
  }
  lStatus=LINK_TOSTART;

  return 0;
}

int CLinkList::SetActiveLinkEnd(int px, int py){
  if ((px!=ActiveLinkX) || (py!=ActiveLinkY)){
    ActiveLinkX = px;
    ActiveLinkY = py;
    return 1;
  }

  return 0;
}

// create a link
int CLinkList::CreateLink(CglModule *From, int FromCh, CglModule *To, int ToCh, int state){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering CreateLink\n");
#endif

  Direction=LINK_DIRECTION_NORMAL;

  Start(From, FromCh);
  Finish(To, ToCh, state);

  return 0;
}

int CLinkList::Reset(){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering Reset\n");
#endif

  TLink *Tmp=Links;

  while (Tmp!=NULL){
    Links=(TLink*) Tmp->Next;
    delete Tmp;
    Tmp=Links;
  };

  return 0;
}

// delete all links connected to the given module
int CLinkList::DeleteModuleLinks(CglModule *mod){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering DeleteModuleLinks\n");
#endif

  TLink *Tmp=Links;
  TLink *Prev=NULL;
  TLink *ToDelete;

  while (Tmp!=NULL){
    if ((Tmp->From==mod) || (Tmp->To==mod)){
      if (Prev==NULL)
	Links=(TLink*) Tmp->Next;
      else
	Prev->Next=Tmp->Next;
      ToDelete=Tmp;
    } else {
      Prev=Tmp;
      ToDelete=NULL;
    }
    Tmp=(TLink*) Tmp->Next;
    if (ToDelete!=NULL) delete ToDelete;
  };

  wxString lm;
  lm << "Links to module '" << mod->GetName().c_str() << "' deleted";
  wxLogMessage(lm);

  return 0;
}

// delete all links connected to the given module at the given channel
int CLinkList::DeleteModuleLinks(CglModule *mod, int ch, int mode){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering DeleteModuleLinks(2)\n");
#endif

  TLink *Tmp=Links;
  TLink *Prev=NULL;
  TLink *ToDelete;

  while (Tmp!=NULL){
    if (  ( (mode==LINK_MODE_DELETE_O) && (Tmp->From==mod) && (Tmp->FromCh==ch) )
	||( (mode==LINK_MODE_DELETE_I) && (Tmp->To  ==mod) && (Tmp->ToCh  ==ch) ) ){
      if (Prev==NULL)
	Links=(TLink*) Tmp->Next;
      else
	Prev->Next=Tmp->Next;
      ToDelete=Tmp;
    } else {
      Prev=Tmp;
      ToDelete=NULL;
    }
    Tmp=(TLink*) Tmp->Next;
    if (ToDelete!=NULL) delete ToDelete;
  };

  wxString lm;
  lm << "Links to module '" << mod->GetName().c_str() << "' at channel " << ch << " deleted";
  wxLogMessage(lm);

  return 0;
}

int CLinkList::ToggleFeedback(CglModule *mod, int ch, int mode){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering ToggleFeedback\n");
#endif

  TLink *Tmp=Links;

  while (Tmp!=NULL){
    if (  ( (mode==LINK_MODE_FEEDBACK_O) && (Tmp->From==mod) && (Tmp->FromCh==ch) )
	||( (mode==LINK_MODE_FEEDBACK_I) && (Tmp->To  ==mod) && (Tmp->ToCh  ==ch) ) ){
      if (Tmp->State==LINK_STATE_FEEDBACK)
	Tmp->State=LINK_STATE_NORMAL;
      else 
	Tmp->State=LINK_STATE_FEEDBACK;
    }
    Tmp=(TLink*) Tmp->Next;
  };

  wxString lm;
  lm << "Toggel links state for module '" << mod->GetName().c_str() << "' at channel " << ch;
  wxLogMessage(lm);

  return 0;
}

// create link section
int CLinkList::CreateLinkSection(FILE *script){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering CreateLinkSection\n");
#endif

  TLink *Tmp;
  int i;

  // calculate number of links
  Tmp=Links;
  i=0;
  while (Tmp!=NULL){
    i++;
    Tmp=(TLink*) Tmp->Next;
  };
  fprintf(script, "%d\n", i);

  wxLogMessage("Saving Link list.");
  // create list
  Tmp=Links;
  i=0;
  while (Tmp!=NULL){
    i++;
    fprintf(script, "%d. %d(%d) -> %d(%d)%s\n", i, Tmp->From->GetIndex(), Tmp->FromCh, Tmp->To->GetIndex(), Tmp->ToCh, (Tmp->State==LINK_STATE_FEEDBACK) ? " FEEDBACK" : "");
    Tmp=(TLink*) Tmp->Next;
  };

  return 0;
}

// read link section
int CLinkList::ReadLinkSection(FILE *script, void *mods, int nmod){
#ifdef DEBUG
  fprintf(stderr, "DBG: linklist: Entering ReadLinkSection\n");
#endif

  int i, j, n, fm, fc, tm, tc, tmp, state;
  TglModuleList *From, *To;
  char Tmp[255], State[255];

  // calculate number of links
  fscanf(script, "%d", &n);
  if (fgets(Tmp, 255, script)==NULL){
    wxLogMessage("Cannot read the project file\n");
    return 1;
  }
  for (i=0;i<n;i++){
    if (fgets(Tmp, 255, script)==NULL){
      wxLogMessage("Cannot read the project file.");
      return 1;
    }
    tmp=sscanf(Tmp, "%d. %d(%d) -> %d(%d)",  &j, &fm, &fc, &tm, &tc);
    // check that the reading has been done correctly
    if (tmp<5){
      wxLogMessage("Cannot read links from the project file.");
      return 1;
    }
    // check if additional state information is given
    memset(State, 0, 255*sizeof(char));
    tmp=sscanf(Tmp, "%d. %d(%d) -> %d(%d) %s", &j, &fm, &fc, &tm, &tc, State);
    j = strlen(State);
    while ((State[j-1]==0x0A) || (State[j-1]==0x0D)){
      j--;
      State[j]=0;
    }
    state = LINK_STATE_NORMAL;
    if (j>0){
      for (tmp=0;tmp<j;tmp++) State[tmp] = toupper(State[tmp]);
      if (strcmp(State, "FEEDBACK")==0){
	wxLogMessage("Link stated as feedback.");
        state=LINK_STATE_FEEDBACK;
      }else{
	wxLogMessage("Error in link section: unknown state.");
        return 1;
      }
    }
    //fscanf(script, "%d. %d(%d) -> %d(%d)", &j, &fm, &fc, &tm, &tc);
    From=(TglModuleList*) mods;
    for (j=0;j<(nmod-fm);j++) From=(TglModuleList*) From->Next;
    To=(TglModuleList*) mods;
    for (j=0;j<(nmod-tm);j++) To=(TglModuleList*) To->Next;
    CreateLink(From->Mod, fc, To->Mod, tc,state);
  }

  return 0;
}
