//--------------------------------------------------------------------//
//  AgALag : Axoris group AL3101 application generator                //
//                                                                    //
//                 _         _    _      _                            //
//                / \____   / \  | |    / \____                       //
//               / _ \  _\ / _ \ | |   / _ \  _\                      //
//              / /_\ \  __ /_\ \| |  / /_\ \  __                     //
//             / _____ \ \ \____ \ |_/______ \ \ \                    //
//            /_/     \_\___\   \_\____|    \_\___\                   //
//                                                                    //
//  File: modorder.cpp                                                //
//  Description: classes dealing with module structure linearization  //
//  Version: 1.0                                                      //
//  Author: Gregor                                                    //
//  e-mail: Gregor.Soyez@swing.be                                     //
//                                                                    //
//  Copyright (C) 2002 Axoris                                         //
//--------------------------------------------------------------------//

#ifndef __AL3101_PROJAL_MODORDER_H__
#define __AL3101_PROJAL_MODORDER_H__

#include "ProjAL.h"
#include "AL3101ModuleHeader.h"

#define MODORDER_OK              0

#define MODORDER_LINK_EXISTS     1
#define MODORDER_INVALID_MODULE  2
#define MODORDER_INVALID_CHANNEL 3
#define MODORDER_CHANNEL_BUSY    4
#define MODORDER_CHANNEL_LOOP    5
#define MODORDER_NO_LINK         6
#define MODORDER_LOOP            7
#define MODORDER_LOOP_ERROR      8

#define MODORDER_USED            10
#define MODORDER_UNUSED          11

#define MODORDER_FILE_NOT_EXISTS 20
#define MODORDER_READ_ERROR      21
#define MODORDER_ERRORS          22

#define MODORDER_INPUT_MODULE    "__AL_INPUT_MODULE"
#define MODORDER_OUTPUT_MODULE   "__AL_OUTPUT_MODULE"

#define LINK_STATE_NORMAL        0
#define LINK_STATE_FEEDBACK      1

// TLink structure
// Description: this represents a plug going into the input channel
// <InputChannel> of a module which index is <moduleindex>.
typedef struct{
  int ModuleIndex;
  void *Module;
  int InputChannel;
  int State;
} TLinkTo;


// TLinkList structure
// Since a module output can be plugged into more than one module,
// we implement dynamic lists of TLinkTo links.
// NOTE: This list is managed by the CLinkList object ... don't use 
//            it directly
typedef struct{
  TLinkTo Plug;
  void *Next;
} TLinkList;


// TProperty definition
// this allows to specify a specific value for a control-parameter 
// in section 2
// NOTE: this list is managed by each module
typedef struct t_property{
  int Index;
  char Value[40];
  t_property *next;
} TProperty;

// CLinkList class
// Class dealing with the TLinkList stack. It should be used by each
// module output channel to specify which modules it has to be connected to.
// It is implemented as a LIFO.
class CLinkList{
 public:
  // ctor & dtor
  CLinkList();
  ~CLinkList();

  // append a link to the stack. The link is given by a pointer to the 
  // destination module (as an index and a void*) and its input channel
  // This returns MODORDER_OK              if successfull
  //              MODORDER_LINK_EXISTS     if link already exists
  //              MODORDER_INVALID_CHANNEL if channel is not valid
  //              MODORDER_CHANNEL_BUSY    if channel is already used
  int Append(int, void*, int, int);

  // Check if the connection list contains a connection to the given module
  // The module is given by its index.
  // This returns MODORDER_OK              if it is the case
  //              MODORDER_INVALID_MODULE  if no such link is found
  int IsPluggedIn(int);

  // place the position pointer to the beginning
  // returns MODORDER_OK
  int Rewind();

  // return the current position link and increment the position
  TLinkTo* GetCurrentLink();

 private:
  TLinkList *LinkList, *Current; // List of links and a pointer to the current place.
  int NbLinks;                   // Number of elements in the list.
};

// CModule class
// This is the class dealing with the properties of a module
class CModule{
 public:
  // ctor & dtor
  CModule();
  ~CModule();

  // Sets the module parameters
  // parameters are the description string or the file module 
  // location and its index
  // Names MODORDER_INPUT_MODULE and MODORDER_OUTPUT_MODULE describe
  // the input and output modules
  int SetParams(char *i_Name, char *i_Label, int i_Index);
  
  // Add a connection to another module
  // The link is given by the output channel, a pointer to the destination 
  // module (as a void*) and its input channel
  // It first call TestLoop to take into account that the link may be a 
  // feedback link. In that case, no update of the structure is required.
  // It calls UpdateStruct for the output module to update the structure's depths
  // This returns MODORDER_OK              if successfull
  //              MODORDER_LINK_EXISTS     if link already exists
  //              MODORDER_INVALID_MODULE  if module index is not valid
  //              MODORDER_INVALID_CHANNEL if channel is not valid
  //              MODORDER_CHANNEL_BUSY    if channel is already used
  //              MODORDER_LOOP            if a loop has been found
  int AddLink(int, CModule*, int, int);

  // update the depth of the current module
  // If the NewDepth is <= current Depth, do nothing and return MODORDER_OK.
  // If the NewDepth is > current Depth, set Depth to NewDepth transmit to all
  // "child" connections" (with NewDepth+1 as argument).
  // The second parameter is used for loop detection. It contains the index of the
  // module where the update starts. If the chain fall back to this module,
  // we have looped.
  // If the current module index is ParentModuleIndex, return MODORDER_LOOP.
  // Else, transmit ParentModuleIndex to all outputs. Returns MODORDER_OK
  // if all of them returns MODORDER_OK, and MODORDER_LOOP if at least one of 
  // them returns MODORDER_LOOP. If all outputs are NULL, return MODORDER_OK.
  // This returns MODORDER_OK     if successfull
  //              MODORDER_LOOP   if a loop has been found
  int UpdateStruct(int NewDepth, int ParentModuleIndex);

  // Check if the input channel passed as parameter is used.
  // returns MODORDER_USED or MODORDER_UNUSED.
  int IsUsed(int);

  // Sets the specified channel to used status.
  // Returns MODORDER_OK;
  int SetUsed(int);

  /*
  ** Add a property (parameter value for section 2)
  ** Parameters are index and value.
  */
  int AddProperty(int, char*);

  int Index;                          // index of the module
  int IsGloballyConnected;            // this is non-null when the module
                                      // is connected to the whole structure
  CLinkList Links[NB_CHANNELS];       // links for each channels
  int Depth;                          // module depth
  char *Name;                         // module name (filename)
  char *Label;                        // module label (title)

  //protected:
  TProperty *Prop;

 private:
  int UseTable;                       // input channel used (binary format)
};

/*
** Fills the table of modules and order them
** Given a file containing the list of modules, we fill the Modules 
** array with the list of modules ORDERED as they should linearily
** appear. The Modules array is supposed to be an allocated vector
** of pointers to CModules objects. A first call to the function
** passing NULL through Modules should only returns the nimber of
** modules through NbModules.
** This function returns MODORDER_OK              if successfull
**                       MODORDER_FILE_NOT_EXISTS if file doesn't exists
**                       MODORDER_READ_ERROR      if a syntax error occurs 
**                       MODORDER_ERRORS          if some structure error is detected
** Outputs are only filled when successful ! 
*/
int CreateStructure(char* FileName, 
		    CModule **Modules, 
		    int *NbModules, 
		    TModuleInfo *pInfo);

#endif

