//--------------------------------------------------------------------//
//  AgALag : Axoris group AL3101 application generator                //
//                                                                    //
//                 _         _    _      _                            //
//                / \____   / \  | |    / \____                       //
//               / _ \  _\ / _ \ | |   / _ \  _\                      //
//              / /_\ \  __ /_\ \| |  / /_\ \  __                     //
//             / _____ \ \ \____ \ |_/______ \ \ \                    //
//            /_/     \_\___\   \_\____|    \_\___\                   //
//                                                                    //
//  File: module.h                                                    //
//  Description: functions dealing with one particular module code    //
//  Version: 1.0                                                      //
//  Author: Gregor                                                    //
//  e-mail: Gregor.Soyez@swing.be                                     //
//                                                                    //
//  Copyright (C) 2002 Axoris                                         //
//--------------------------------------------------------------------//

#ifndef __PROJ_AL_MODULE_H__
#define __PROJ_AL_MODULE_H__

#include "ProjAL.h"
#include "modorder.h"
#include "AL3101ModuleHeader.h"
#include "replace.h"
#include <stdio.h>

#define MODULE_OK               0
#define MODULE_MISSING_SECTION  1
#define MODULE_FILE_ERROR       2
#define MODULE_ADD_ERROR        3
#define MODULE_WRONG_STRUCTURE  4
#define MODULE_NO_TMP_LEFT      5
#define MODULE_SYNTAX_ERROR     6
#define MODULE_NOT_PROPERTY     7

#define MODULE_SILENT          10
#define MODULE_VERBOSE         11

/*
** This class defines a AL3101 code module
** All the informations about how to place
** the module into the structure are contained 
** into the CModule member
*/ 
class CCodeModule: 
public CModule,
public CReplacementList
{
 public:
  // ctor & dtor
  CCodeModule();
  ~CCodeModule();

  TModuleInfo info;

  int CatchedTmpIO[2*NB_CHANNELS];        // contain the temporary addresses catched by a module (8 maximum)
  long int SectionsOffset[NB_SECTIONS+1]; // position of section in source file
                                          // filled by CheckSections

  /* 
  ** browse the source code and search for sections headers
  ** This fills the SectionsOffset table with sarting points 
  ** of every section in the file
  ** It returns MODULE_OK on success, MODULE_FILE_ERROR if 
  ** an error occurs with file and MODULE_MISSING_SECTION
  ** if some section seems to miss.
  */
  int CheckSections();

  /*
  ** Retreive the number of cycles used by the module 
  ** and the memory needed. It also tells wether (1) 
  ** or not (0) the module needs a delay line
  ** These are given by 
  ** ; Cycles      : ? cycles (-1=unknownw)
  ** ; Memory      : ? words (without init)
  ** ; Delay Line  :	no/yes (remove wrong one);
  ** in the module header.
  ** The last function initialize the memory offset for the module.
  */
  int GetCycles();
  int GetMemoryNeeds();
  int GetDelayLine();
  int SetMemOffset(int i_offset);
  int GetMemOffset();
  //int PrintMemOffset(FILE *flux);

  /*
  ** All the outputs which are not directly mapped on the OUTPUT module
  ** have to pass through a DIR? temporary variable. Moreover, all
  ** inputs coming from module different from line input have to
  ** be replaced by the DIR? output of the preceeding module.
  ** This function creates the table of these replacements . It
  ** uses the tools in iomapping.h. It returns MODULE_OK when 
  ** successful, MODULE_NO_DIR_LEFT if there are not enough free
  ** DIR? temporary variables and MODULE_ADD_ERROR if the link 
  ** for output could not be added.
  ** The outputs can be modified internally and the inputs are 
  ** modified by calling the second function for the next module
  ** from the first one. That second function return MODULE_ADD_ERROR 
  ** Iin case of troubles and MODULE_OK if everything's fine.
  */
  int CreateTemporaryIO();
  int AddTemporaryInput(int, int);

  /*
  ** Append the requested section to the given stream
  */
  int AppendSectionToStream(int, FILE*);
  int InitRemoved();

  /*
  ** Append the requested section to the given stream
  */
  int UpdateProperty(int, char*);

 private:
  char DefStr[80];//, MemDefStr[160];
  int NbCycles, MemNeeds;                 // Number of cycles and memory used
  int IsDelayLine;                        // 1 if delay line, 0 elsewhere
  int MemoryOffset;                       // Memory Offset

  /* 
  ** Every Application & mModule consists in 7 sections :
  **   SEC1 Control parameters
  **   SEC2 Initial values of control parameters (S3.24 format)
  **   SEC3 Constants definitions
  **   SEC4 Memory variables definition
  **   SEC5 Initial values of memory variables
  **   SEC6 Initialization
  **   SEC7 Module code
  ** The following functions make the insertion work typical for each section
  ** They return MODULE_OK on success and MODULE_SYNTAX_ERROR in
  ** case of trouble(s).
  */
  int DealCtrlParam(FILE*, FILE*);
  int DealInitVal(FILE*);
  int DealCstDef(FILE*);
  int DealMemDef(FILE*, FILE*);
  int DealInitMem(FILE*);
  int DealInitSec(FILE*);
  int DealCode(FILE*);
};

#endif
