// FileIO_DEC.cpp: implementation of the DEC File functions.
//
//////////////////////////////////////////////////////////////////////

#include "FileIO_DEC.h"
#include "FileIO_DEC_internal.h"

#include <stdlib.h>		// for the malloc function
#include <string.h>		// for the string routines in error functions


//////////////////////////////////////////////////////////////////////
// init & clean functions
//////////////////////////////////////////////////////////////////////
INT32 FileIO_DEC_Init(void **pInstance, CHAR *FileName, INT32 Mode, FileIO_SndCfg *cfg, INT32 *LeftSampleMemory, INT32 *RightSampleMemory)
{
	FileIO_DEC_Struct *pInst;

	// checking and getting file characteristics
	if (cfg->NumberOfBytesInSample != 2) return FILEIO_TYPE_NOT_SUPPORTED;
	if (cfg->NumberOfChannels != 1) return FILEIO_TYPE_NOT_SUPPORTED;
	// moreover checking if The number of channel wanted is not OK with what available in the file
	//if (((RightSampleMemory == NULL) && (cfg->NumberOfChannels == 2)) || ((RightSampleMemory != NULL) && (cfg->NumberOfChannels != 2))) return FILEIO_WRONG_CHANNEL_NUMBER;
	// this test is not required as we only support mono here

	// Allocate the space for the instance structure
	pInst = new FileIO_DEC_Struct;
	if (pInst == NULL) return FILEIO_CANNOT_ALLOCATE_INSTANCE;
	*pInstance = (void *)pInst;

	pInst->Mode = Mode;
	pInst->CurrentPos = 0;

	// open the file
	pInst->SamplingFrequency = cfg->SamplingFrequency;		// The sampling frequency of the file
	if (pInst->Mode != 0)
	{
		pInst->file = fopen(FileName, "w+t");
	}
	else
	{
		pInst->file = fopen(FileName, "rt");
	}
	if (pInst->file == NULL)
	{
		delete pInst;
		return FILEIO_CANNOT_OPEN_FILE;
	}
	pInst->Loop = 0;					// By default, no looping


	// do the DSP memory related stuff initialization
	pInst->LeftData = LeftSampleMemory;			// Pointer to the left sample memory
	pInst->RightData = RightSampleMemory;		// Pointer to the right sample memory
	
	return FILEIO_NOERROR;	// no problem occurs
}

INT32 FileIO_DEC_Clean(void *pInstance)
{
	FileIO_DEC_Struct *pInst = (FileIO_DEC_Struct *)pInstance;

	// closing the file	
	if (pInst->file != NULL) if (fclose(pInst->file) != 0) return FILEIO_CANNOT_CLOSE_FILE;

	// freeing the instance structure
	if (pInst != NULL) delete pInst;

	return FILEIO_NOERROR;
}


//////////////////////////////////////////////////////////////////////
// "processing" functions
//////////////////////////////////////////////////////////////////////
INT32 FileIO_DEC_ReadSample(void *pInstance)
{
	FileIO_DEC_Struct *pInst = (FileIO_DEC_Struct *)pInstance;
	INT32 Value;

	// Check if we try to read in a file which has been opened in that mode
	if (pInst->Mode != 0)
	{
		return FILEIO_WRONG_FILE_MODE;			// Try to read in a file that is in write mode
	}
	
	
	// Check if data is still available in the buffer
	INT32 result = fscanf(pInst->file, "%d\n", &Value);
	if (result != 1)
	{
		if (!feof(pInst->file))
		{
			return FILEIO_READ_ERROR;			// A problem occurs in ReadFile
		}
		else
		{
			if (pInst->Loop) fseek(pInst->file, 0, SEEK_SET);
			else return FILEIO_EOF_REACHED;		// End of file reached, reset to beginning of file
		}
	}

	// let's now transfer data from the buffer to the memory
	*(pInst->LeftData) = Value;					// copy left/mono channel to memory
	
	pInst->CurrentPos = pInst->CurrentPos + 1;

	return FILEIO_NOERROR;	// no problem occurs
}

	
INT32 FileIO_DEC_WriteSample(void *pInstance)
{
	FileIO_DEC_Struct *pInst = (FileIO_DEC_Struct *)pInstance;
	INT32 Value;

	// Check if we try to read in a file which has been opened in that mode
	if (pInst->Mode == 0)
	{
		return FILEIO_WRONG_FILE_MODE;				// Try to write in a file that is in read mode
	}

	// let's transfer data from the memory to the buffer
	Value = *(pInst->LeftData);

	INT32 result = fprintf(pInst->file, "%d\n", Value);
	if (result < 0)
	{
		return FILEIO_WRITE_ERROR;					// A problem occurs in WriteFile
	}
	
	pInst->CurrentPos = pInst->CurrentPos + 1;

	return FILEIO_NOERROR;	// no problem occurs
}



//////////////////////////////////////////////////////////////////////
// file function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_DEC_GetPosition(void *pInstance, INT32 *CurrentPosition)
{
	FileIO_DEC_Struct *pInst = (FileIO_DEC_Struct *)pInstance;

	*CurrentPosition = pInst->CurrentPos;

	return FILEIO_NOERROR;
}

INT32 FileIO_DEC_GetFileLength(void *pInstance, INT32 *FileSize)
{
	CHAR str[256];
	INT32 result;
	INT32 CurrentPosition = 0;
	INT32 size = 0;

	FileIO_DEC_Struct *pInst = (FileIO_DEC_Struct *)pInstance;

	// Get the current position in the file and return if impossible
	CurrentPosition = ftell(pInst->file);
	if (CurrentPosition == -1) return FILEIO_CANNOT_GET_POSITION;

	// Seek to the end of file
	result = fseek(pInst->file, 0, SEEK_SET);
	if (result != 0) return FILEIO_PROBLEM_WHILE_SEEKING;

	while(1)
	{
		if (fgets(str, 255, pInst->file) == NULL) break;
		else size++;
	}

	//  Seek to the position where the file was before
	result = fseek(pInst->file, CurrentPosition, SEEK_SET);
	if (result != 0) return FILEIO_PROBLEM_WHILE_SEEKING;


	*FileSize = size;


	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// control function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_DEC_SetLoopingMode(void *pInstance, INT32 Looping)
{
	FileIO_DEC_Struct *pInst = (FileIO_DEC_Struct *)pInstance;
	pInst->Loop = Looping;

	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// "text" function(s)
//////////////////////////////////////////////////////////////////////
void FileIO_DEC_GetLastErrorString(CHAR *str, INT32 LastErrorCode, INT32 MaxNumberOfCharacters)
{
	if ((LastErrorCode & FILEIO_BASE_ERR) == FILEIO_BASE_ERR)		// check if the error is a FileIO error
	{
		if ((LastErrorCode & FILEIO_GEN_BASE_ERR) == FILEIO_GEN_BASE_ERR)
		{
			FileIO_GetLastErrorString(str, LastErrorCode, MaxNumberOfCharacters);
			return;
		}
		else if ((LastErrorCode & FILEIO_DEC_BASE_ERR) == FILEIO_DEC_BASE_ERR)
		{
			// do nothing here, wait for the switch below
		}
		else
		{
			strncpy(str, "Error type not known", MaxNumberOfCharacters - 1);
		}
	}
	else															// this error is not a FileIO error
	{
		strncpy(str, "Not a FileIO error", MaxNumberOfCharacters - 1);
		return;
	}


	switch(LastErrorCode)
	{
		case FILEIO_DEC_NOERROR:
			strncpy(str, "DEC: No error occurred", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "DEC: Error not known", MaxNumberOfCharacters - 1);
			break;
	}
	return;
}



void FileIO_DEC_GetHelp(CHAR *str, INT32 LineNumber, INT32 MaxNumberOfCharacters)
{
	switch(LineNumber)
	{
		case 0:
			strncpy(str, "DEC files are read by these functions.", MaxNumberOfCharacters - 1);
			break;
		case 1:
			strncpy(str, "DEC files are raw data encoded in Little Endian format i.e. LSB,MSB", MaxNumberOfCharacters - 1);
			break;
		case 2:
			strncpy(str, "There's no support for 8 bits files and for more than 4GB files.", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "DEC file functions", MaxNumberOfCharacters - 1);
	}
}
