/********************************************************************/
/*                                                                  */
/*                        A X O R I S                               */
/*                                                                  */
/*                Alesis AL3101/2 DSP Simulator                     */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/* Origin Date    : June 2001                                       */
/* Project Version: v0.1                                            */
/* FileName       : AL3101_IOs.c                                    */
/* File Version   : v0.1                                            */
/* Author         : Fabian Duchesne                                 */
/* Last Modif     : September 2001                                  */
/*                                                                  */
/********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "./../MessageMacros.h"
#include "AL3101_Core.h"
#include "AL3101_IOs.h"
#include "./../CommandLine/Commands.h"
#include "./../1K_func/AL3101Format.h"
#include "./../1K_func/AL3101AsmDisasmUtils.h"
#include "./../Debug.h"
#include "./../Error.h"
#include "./../HelperFunctions.h"
// includes for the FileIO
#include "./../FileIO/HEX/FileIO_HEX.h"
#include "./../FileIO/DEC/FileIO_DEC.h"
#include "./../FileIO/DAT/FileIO_DAT.h"
#include "./../FileIO/PCM/FileIO_PCM.h"
#include "./../FileIO/WAV/FileIO_WAV.h"
// includes for the Tools
#include "./../Tools/Tool_AmplitudeSweep.h"
#include "./../Tools/Tool_FrequencyResponse.h"
// include for the generators
#include "./../Tools/Gen_SIN.h"
#include "./../Tools/Gen_WNO.h"
#include "./../Tools/Gen_TRI.h"


/********************************************************************/
/*                     I O s   L I S T                              */
/********************************************************************/
INT32 NumOfInputModesInList = 4;
INT32 NumOfOutputModesInList = 3;
INT32 NumOfSignalModesInList = 3;
INT32 NumOfTraceModesInList = 2;

AL3101IoMode AL3101InputModesList[] = {
    { "file"      , 4 , IN_FILE       , InFile_Init       , "" },
	{ "gen"		  , 3 , IN_GENERATOR  , InGen_Init		  , "" },
	{ "wavedev"   , 7 , IN_LINE       , InWaveDev_Init    , "" },
	{ ""          , 0 , 0             , 0                 , "" }
};

AL3101IoMode AL3101OutputModesList[] = {
	{ "file"      , 4 , OUT_FILE      , OutFile_Init      , "" },
	{ "wavedev"   , 7 , OUT_LINE      , OutWaveDev_Init   , "" },
	{ ""          , 0 , 0             , 0                 , "" }
};

AL3101IoMode AL3101SignalModesList[] = {
    { "file"      , 4 , IN_FILE       , InFile_Init       , "" },
	{ "gen"		  , 3 , IN_GENERATOR  , InGen_Init		  , "" },
	{ ""          , 0 , 0             , 0                 , "" }
};

AL3101IoMode AL3101TraceModesList[] = {
	{ "file"      , 4 , OUT_FILE      , OutFile_Init      , "" },
	{ ""          , 0 , 0             , 0                 , "" }
};

/************************************************************/
/*****          I N P U T   F U N C T I O N S           *****/
/************************************************************/

/************************************************************/
/* NAME       :												*/
/* DESCRIPTION:												*/
/* INPUT      :												*/
/* OUTPUT     :												*/
/************************************************************/
INT32 InFile_Init(AL3101* pDspInstance, CHAR* Options, IOTable *InputInstance, INT32 *InputInstanceIndex)
{
	FileIO_SndCfg cfg;
	CHAR FileName[CMD_MAX_LEN];
	CHAR TempStr[CMD_MAX_LEN];
	CHAR OffsetString[CMD_MAX_LEN];
	INT32 Offset;
	INT32 *LeftSampleMemory, *RightSampleMemory = NULL;
	INT32 RegMemSpecifier;
	INT32 TempSamplingFrequency;
	INT32 Result;
	INT32 index;

	DBG_PRINT("In InFile_Init\n");

	// Saving the command line for future use
	strcpy(InputInstance[*InputInstanceIndex].CmdLine, Options);

	// Remove the "file" keyword from the string. FileName will be overwritten with the correct string later on
	GetNextWord(FileName, Options);

	// Get the file name
	GetNextFilenameWord(FileName, Options);
	for(index = strlen(FileName); index >= 0; index--) FileName[index] = tolower(FileName[index]);

	// Extract the extension from the file name
	CHAR *pCharTemp = strrchr(FileName, '.');		// search for the '.'
	if (pCharTemp == NULL)
	{
		ERROR("Unable to determine the extension of the file.\n");
		return CMD_FILE_CANNOT_RETRIEVE_EXTENSION;
	}
	pCharTemp++;									// skip the '.' and go to the extension

	// put the rest of the string in lower case
 	String_ToLower_String(Options, Options);

	// Get the sampling frequency value
	GetNextWord(TempStr, Options);
	TempSamplingFrequency = atoi(TempStr);

	// Checking that the sampling frequency is OK with the previously initialized inputs
	if (TempSamplingFrequency != 0)						// If given sampling frequency is 0, do not take it in account
	{
		// first check the "absolute" validity of the value regarding the DSP capabilities
		if ((TempSamplingFrequency < 30000) || (TempSamplingFrequency > 50000))
		{
			WARNING("Warning: the recommended sampling frequency should be between 30kHz and 50kHz.\n");
			WARNING_1(" The sampling frequency required is %d.", TempSamplingFrequency);
			WARNING(" Simulation will go on but will perhaps not represent the real hardware operation.\n");
		}
		// then check the "relative" validity of the value regarding the inputs already set
		if (pDspInstance->SamplingFrequency == -1) pDspInstance->SamplingFrequency = TempSamplingFrequency;		// The first time, the sampling frequency used in the program is set
		else
		{
			if (pDspInstance->SamplingFrequency != TempSamplingFrequency)							// Next times, test is done
			{
				ERROR_1("The sampling frequency used in file %s\n  is different from the sampling frequency used in other inputs\n", FileName);
				ERROR(" Please check that all sampling frequencies are the same.\n");
				return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
			}
		}
	}
	
	// Get the left offset value
	// Get the reg/mem specifier or the left offset value
	GetNextWord(OffsetString, Options);
	if (OffsetString == NULL)
	{
		ERROR("Left/Mono offset not specified.\n");
		return CMD_FILE_WRONG_ARGUMENT;
	}
	GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
	switch (RegMemSpecifier)
	{
	case -1:		// unknown stuff
		ERROR("Unknown register or memory specifier. Giving up...\n");
		return CMD_FILE_WRONG_ARGUMENT;
	case 0:			// this is a number, assuming "In" memory use
		LeftSampleMemory = pDspInstance->State.In;
		Offset = atoi(OffsetString);
		break;
	default:		// there is a register or memory specifier
		GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &LeftSampleMemory);
		if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
		{
			GetNextWord(OffsetString, Options);
			if (OffsetString == NULL)
			{
				ERROR("Left/Mono location not specified.\n");
				return CMD_FILE_WRONG_ARGUMENT;
			}
			Offset = atoi(OffsetString);
		}
		else Offset = 0;
	}
	// Add the offset to the base address.
	LeftSampleMemory = LeftSampleMemory + Offset;


	// Get the right offset value
	// Get the reg/mem specifier or the left offset value
	GetNextWord(OffsetString, Options);
	if (OffsetString == NULL)
	{
		ERROR("Left/Mono offset not specified.\n");
		return CMD_FILE_WRONG_ARGUMENT;
	}
	else if (strcmp(OffsetString, "") == 0)
	{
		Offset = -1;
		RightSampleMemory = NULL;
	}
	else
	{
		GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
		switch (RegMemSpecifier)
		{
		case -1:		// unknown stuff
			ERROR("Unknown register or memory specifier. Giving up...\n");
			return CMD_FILE_WRONG_ARGUMENT;
		case 0:			// this is a number, assuming "In" memory use
			RightSampleMemory = pDspInstance->State.In;
			Offset = atoi(OffsetString);
			break;
		default:		// there is a register or memory specifier
			GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &RightSampleMemory);
			if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
			{
				GetNextWord(OffsetString, Options);
				if (OffsetString == NULL)
				{
					ERROR("Right location not specified.\n");
					return CMD_FILE_WRONG_ARGUMENT;
				}
				Offset = atoi(OffsetString);
			}
			else Offset = 0;
		}
		// Add the offset to the base address.
		RightSampleMemory = RightSampleMemory + Offset;
	}


	// making the file type specific things
	if (strcmp(pCharTemp, "wav") == 0)			// The input file is a .wav file
	{
		// Initialize with the FILEIO wav function
		Result = FileIO_WAV_Init(&InputInstance[*InputInstanceIndex].pInstance, FileName, 0, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Checking that the sampling frequency is OK with the previously initialized inputs
		if ((cfg.SamplingFrequency < 30000) || (cfg.SamplingFrequency > 50000))
		{
			WARNING("Warning: the recommended sampling frequency should be between 30kHz and 50kHz.\n");
			WARNING_1(" The sampling frequency required is %d.\n", cfg.SamplingFrequency);
			WARNING(" Simulation will go on but will perhaps not represent the hardware operations.\n");
		}
		if (pDspInstance->SamplingFrequency == -1) pDspInstance->SamplingFrequency = cfg.SamplingFrequency;		// The first time, the sampling frequency used in the program is set
		else
		{
			if (pDspInstance->SamplingFrequency != cfg.SamplingFrequency)							// Next times, test is done
			{
				ERROR_1("The sampling frequency used in file %s\n  is different from the sampling frequency used in other inputs\n", FileName);
				ERROR(" Please check that all sampling frequencies are the same.\n");
				FileIO_WAV_Clean(InputInstance[*InputInstanceIndex].pInstance);
				return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
			}
		}
		// Initialize table
		InputInstance[*InputInstanceIndex].f_Action = &FileIO_WAV_ReadSample;
		InputInstance[*InputInstanceIndex].f_Clean = &FileIO_WAV_Clean;
		InputInstance[*InputInstanceIndex].f_GetPosition = &FileIO_WAV_GetPosition;
		InputInstance[*InputInstanceIndex].f_GetLength = &FileIO_WAV_GetFileLength;
		InputInstance[*InputInstanceIndex].f_SetLooping = &FileIO_WAV_SetLoopingMode;
		*InputInstanceIndex = *InputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "pcm") == 0)		// The input file is a .pcm file
	{
		// Initialize the cfg structure
		if (TempSamplingFrequency == -1)
		{
			ERROR("The sampling frequency cannot be reused since it has not been defined yet.\n");
			ERROR("  Please define the sampling frequency with another value than 0.\n");
			ERROR("  Or use a file containing the sampling frequency in a header (like wav files) before.\n");
			return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
		}
		else cfg.SamplingFrequency = TempSamplingFrequency;
		if (RightSampleMemory == NULL) cfg.NumberOfChannels = 1;
		else cfg.NumberOfChannels = 2;							
		cfg.NumberOfBytesInSample = 2;							// the number of bytes per sample is always 2 => 16 bits file
		// Initialize with the FILEIO pcm function
		Result = FileIO_PCM_Init(&InputInstance[*InputInstanceIndex].pInstance, FileName, 0, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		InputInstance[*InputInstanceIndex].f_Action = &FileIO_PCM_ReadSample;
		InputInstance[*InputInstanceIndex].f_Clean = &FileIO_PCM_Clean;
		InputInstance[*InputInstanceIndex].f_GetPosition = &FileIO_PCM_GetPosition;
		InputInstance[*InputInstanceIndex].f_GetLength = &FileIO_PCM_GetFileLength;
		InputInstance[*InputInstanceIndex].f_SetLooping = &FileIO_PCM_SetLoopingMode;
		*InputInstanceIndex = *InputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "dat") == 0)		// The input file is a .dat file
	{
		// Initialize the cfg structure
		if (TempSamplingFrequency == -1)
		{
			ERROR("The sampling frequency cannot be reused since it has not been defined yet.\n");
			ERROR("  Please define the sampling frequency with another value than 0.\n");
			ERROR("  Or use a file containing the sampling frequency in a header (like wav files) before.\n");
			return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
		}
		else cfg.SamplingFrequency = TempSamplingFrequency;
		cfg.NumberOfChannels = 1;					// forcing the number of channels to 1
		if (RightSampleMemory != NULL)				// explaining that left only is handled
		{
			WARNING("Dat files only support mono.\nRight channel will be ignored.\n");
		}
		cfg.NumberOfBytesInSample = 2;							// the number of bytes per sample is always 2 => 16 bits file
		// Initialize with the FILEIO dat function
		Result = FileIO_DAT_Init(&InputInstance[*InputInstanceIndex].pInstance, FileName, 0, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		InputInstance[*InputInstanceIndex].f_Action = &FileIO_DAT_ReadSample;
		InputInstance[*InputInstanceIndex].f_Clean = &FileIO_DAT_Clean;
		InputInstance[*InputInstanceIndex].f_GetPosition = &FileIO_DAT_GetPosition;
		InputInstance[*InputInstanceIndex].f_GetLength = &FileIO_DAT_GetFileLength;
		InputInstance[*InputInstanceIndex].f_SetLooping = &FileIO_DAT_SetLoopingMode;
		*InputInstanceIndex = *InputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "dec") == 0)		// The input file is a .dec file
	{
		// Initialize the cfg structure
		if (TempSamplingFrequency == -1)
		{
			ERROR("The sampling frequency cannot be reused since it has not been defined yet.\n");
			ERROR("  Please define the sampling frequency with another value than 0.\n");
			ERROR("  Or use a file containing the sampling frequency in a header (like wav files) before.\n");
			return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
		}
		else cfg.SamplingFrequency = TempSamplingFrequency;
		cfg.NumberOfChannels = 1;					// forcing the number of channels to 1
		if (RightSampleMemory != NULL)				// explaining that left only is handled
		{
			WARNING("Int files only support mono.\nRight channel will be ignored.\n");
		}
		cfg.NumberOfBytesInSample = 2;							// the number of bytes per sample is always 2 => 16 bits file
		// Initialize with the FILEIO dec function
		Result = FileIO_DEC_Init(&InputInstance[*InputInstanceIndex].pInstance, FileName, 0, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		InputInstance[*InputInstanceIndex].f_Action = &FileIO_DEC_ReadSample;
		InputInstance[*InputInstanceIndex].f_Clean = &FileIO_DEC_Clean;
		InputInstance[*InputInstanceIndex].f_GetPosition = &FileIO_DEC_GetPosition;
		InputInstance[*InputInstanceIndex].f_GetLength = &FileIO_DEC_GetFileLength;
		InputInstance[*InputInstanceIndex].f_SetLooping = &FileIO_DEC_SetLoopingMode;
		*InputInstanceIndex = *InputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "hex") == 0)		// The input file is a .hex file
	{
		// Initialize the cfg structure
		if (TempSamplingFrequency == -1)
		{
			ERROR("The sampling frequency cannot be reused since it has not been defined yet.\n");
			ERROR("  Please define the sampling frequency with another value than 0.\n");
			ERROR("  Or use a file containing the sampling frequency in a header (like wav files) before.\n");
			return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
		}
		else cfg.SamplingFrequency = TempSamplingFrequency;
		cfg.NumberOfChannels = 1;					// forcing the number of channels to 1
		if (RightSampleMemory != NULL)				// explaining that left only is handled
		{
			WARNING("Hex files only support mono.\nRight channel will be ignored.\n");
		}
		cfg.NumberOfBytesInSample = 2;							// the number of bytes per sample is always 2 => 16 bits file
		// Initialize with the FILEIO hex function
		Result = FileIO_HEX_Init(&InputInstance[*InputInstanceIndex].pInstance, FileName, 0, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		InputInstance[*InputInstanceIndex].f_Action = &FileIO_HEX_ReadSample;
		InputInstance[*InputInstanceIndex].f_Clean = &FileIO_HEX_Clean;
		InputInstance[*InputInstanceIndex].f_GetPosition = &FileIO_HEX_GetPosition;
		InputInstance[*InputInstanceIndex].f_GetLength = &FileIO_HEX_GetFileLength;
		InputInstance[*InputInstanceIndex].f_SetLooping = &FileIO_HEX_SetLoopingMode;
		*InputInstanceIndex = *InputInstanceIndex + 1;
	}
	else										// The input file extension is not known
	{
		ERROR_1("Sorry, the file extension: %s is not known.\n", pCharTemp);
		return CMD_FILE_UNKNOWN_EXTENSION;
	}
	
	DBG_PRINT("Out InFile\n");

	return CMD_FILE_NOERROR;
};

// Helper function for InGen_Init
//  Gets amplitude and offset from the string
INT32 GetAmplitudeAndOffset(CHAR *Options, float *Amplitude, float *Offset)
{
	CHAR TempStr[CMD_MAX_LEN];
	INT32 Value;

	// Get first the amplitude
	GetNextWord(TempStr, Options);
	String_ToUpper_String(TempStr, TempStr);
	if (ExtractHexa(TempStr, &Value, 24) != 0)
	{
		ERROR_1("Error in converting value: %s\n", TempStr);
		return -1;
	}
	ExtendTo32Bits(&Value);
	*Amplitude = (float)Value / 16777216.0f;
	if (*Amplitude < -8.0f)
	{
		WARNING("Amplitude cannot exceed -8.0. Setting it to -8,0...\n");
		*Amplitude = -8.0f;
	}
	if (*Amplitude > 8.0f)
	{
		WARNING("Amplitude cannot exceed 8.0. Setting it to 8,0... \n");
		*Amplitude = 8.0f;
	}

	// Get then the offset
	GetNextWord(TempStr, Options);
	if (strcmp(TempStr, "") != 0)
	{
		String_ToUpper_String(TempStr, TempStr);
		if (ExtractHexa(TempStr, &Value, 24) != 0)
		{
			ERROR_1("Error in converting value: %s\n", TempStr);
			return -1;
		}
		ExtendTo32Bits(&Value);
		*Offset = (float)Value / 16777216.0f;
	}
	else *Offset = 0.0f;

	return 0;
}

/************************************************************/
/* NAME       :												*/
/* DESCRIPTION:												*/
/* INPUT      :												*/
/* OUTPUT     :												*/
/************************************************************/
INT32 InGen_Init(AL3101* pDspInstance, CHAR* Options, IOTable *InputInstance, INT32 *InputInstanceIndex)
{
	CHAR GenName[CMD_MAX_LEN];
	CHAR TempStr[CMD_MAX_LEN];
	CHAR OffsetString[CMD_MAX_LEN];
	INT32 Offset;
	INT32 *SampleMemory;
	INT32 RegMemSpecifier;
	INT32 TempSamplingFrequency;
	INT32 Result;
	INT32 index;

	DBG_PRINT("In InGen_Init\n");

	// Saving the command line for future use
	strcpy(InputInstance[*InputInstanceIndex].CmdLine, Options);

	// Remove the "gen" keyword from the string. FileName will be overwritten with the correct string later on
	GetNextWord(GenName, Options);

	// Get the type name
	GetNextWord(GenName, Options);
	for(index = strlen(GenName); index >= 0; index--) GenName[index] = tolower(GenName[index]);

	// put the rest of the string in lower case
 	String_ToLower_String(Options, Options);

	// Get the sampling frequency value
	GetNextWord(TempStr, Options);
	TempSamplingFrequency = atoi(TempStr);

	// Checking that the sampling frequency is OK with the previously initialized inputs
	if (TempSamplingFrequency != 0)						// If given sampling frequency is 0, do not take it in account
	{
		// first check the "absolute" validity of the value regarding the DSP capabilities
		if ((TempSamplingFrequency < 30000) || (TempSamplingFrequency > 50000))
		{
			WARNING("Warning: the recommended sampling frequency should be between 30kHz and 50kHz.\n");
			WARNING_1(" The sampling frequency required is %d.", TempSamplingFrequency);
			WARNING(" Simulation will go on but will perhaps not represent the real hardware operation.\n");
		}
		// then check the "relative" validity of the value regarding the inputs already set
		if (pDspInstance->SamplingFrequency == -1) pDspInstance->SamplingFrequency = TempSamplingFrequency;		// The first time, the sampling frequency used in the program is set
		else
		{
			if (pDspInstance->SamplingFrequency != TempSamplingFrequency)							// Next times, test is done
			{
				ERROR_1("The sampling frequency used by this generator %s\n  is different from the sampling frequency used in other inputs\n", GenName);
				ERROR(" Please check that all sampling frequencies are the same.\n");
				return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
			}
		}
	}
	
	// Get the sample offset value
	// Get the reg/mem specifier of the sample offset value
	GetNextWord(OffsetString, Options);
	if (OffsetString == NULL)
	{
		ERROR("Mono offset not specified.\n");
		return CMD_FILE_WRONG_ARGUMENT;
	}
	GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
	switch (RegMemSpecifier)
	{
	case -1:		// unknown stuff
		ERROR("Unknown register or memory specifier. Giving up...\n");
		return CMD_FILE_WRONG_ARGUMENT;
	case 0:			// this is a number, assuming "In" memory use
		SampleMemory = pDspInstance->State.In;
		Offset = atoi(OffsetString);
		break;
	default:		// there is a register or memory specifier
		GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &SampleMemory);
		if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
		{
			GetNextWord(OffsetString, Options);
			if (OffsetString == NULL)
			{
				ERROR("Mono location not specified.\n");
				return CMD_FILE_WRONG_ARGUMENT;
			}
			Offset = atoi(OffsetString);
		}
		else Offset = 0;
	}
	// Add the offset to the base address.
	SampleMemory = SampleMemory + Offset;


	if (strcmp(GenName, "wnoise") == 0)		// The generator should generate white noise
	{
		Generator_WhiteNoiseCfg cfg;
		float Amplitude;
		float Offset;

		// Get the amplitude and offset
		if (GetAmplitudeAndOffset(Options, &Amplitude, &Offset) != 0) return CMD_FILE_WRONG_ARGUMENT;

		cfg.Amplitude = Amplitude;
		cfg.Offset = Offset;

		// Initialize with the Gen_WNO function
		Result = Gen_WNO_Init(&InputInstance[*InputInstanceIndex].pInstance, &cfg, SampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing generator: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		InputInstance[*InputInstanceIndex].f_Action = &Gen_WNO_ReadSample;
		InputInstance[*InputInstanceIndex].f_Clean = &Gen_WNO_Clean;
		InputInstance[*InputInstanceIndex].f_GetPosition = &Gen_WNO_GetPosition;
		InputInstance[*InputInstanceIndex].f_GetLength = &Gen_WNO_GetFileLength;
		InputInstance[*InputInstanceIndex].f_SetLooping = &Gen_WNO_SetLoopingMode;
		*InputInstanceIndex = *InputInstanceIndex + 1;
	}
	else
	{
		float GenFrequency;
		float Amplitude;
		float Offset;

		// Get the operating frequency in Hz
		GetNextWord(TempStr, Options);
		sscanf(TempStr, "%f", &GenFrequency);
		if (GenFrequency < 0.0f)
		{
			WARNING("Generator frequency cannot be negative. Setting to 100Hz instead...\n");
			GenFrequency = 100.0f;
		}
		if (GenFrequency > (TempSamplingFrequency / 2))
		{
			WARNING("Generator frequency should not exceed Fs/2. Setting to Fs/2 instead...\n");
			GenFrequency = TempSamplingFrequency / 2;
		}

		// Get the amplitude and offset
		if (GetAmplitudeAndOffset(Options, &Amplitude, &Offset) != 0) return CMD_FILE_WRONG_ARGUMENT;

		// making the file type specific things
		if (strcmp(GenName, "sin") == 0)			// The generator should generate a sine wave
		{
			Generator_SineCfg cfg;
			float OriginalPhase = 0.0f;

			// Get the original phase in rad
			GetNextWord(TempStr, Options);
			sscanf(TempStr, "%f", &OriginalPhase);

			cfg.SineFrequency = GenFrequency;
			cfg.Amplitude = Amplitude;
			cfg.Offset = Offset;
			cfg.OriginalPhase = OriginalPhase;

			cfg.SamplingFrequency = TempSamplingFrequency;

			// Initialize with the Gen_SIN function
			Result = Gen_SIN_Init(&InputInstance[*InputInstanceIndex].pInstance, &cfg, SampleMemory);
			if (Result != FILEIO_NOERROR)
			{
				FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
				ERROR_1("Error in initializing generator: %s\n", OffsetString);
				return CMD_FILE_CANNOT_INITIALIZE_FILE;
			}
			// Initialize table
			InputInstance[*InputInstanceIndex].f_Action = &Gen_SIN_ReadSample;
			InputInstance[*InputInstanceIndex].f_Clean = &Gen_SIN_Clean;
			InputInstance[*InputInstanceIndex].f_GetPosition = &Gen_SIN_GetPosition;
			InputInstance[*InputInstanceIndex].f_GetLength = &Gen_SIN_GetFileLength;
			InputInstance[*InputInstanceIndex].f_SetLooping = &Gen_SIN_SetLoopingMode;
			*InputInstanceIndex = *InputInstanceIndex + 1;
		}
		else if (strcmp(GenName, "tri") == 0)		// The generator should generate triangle
		{
			Generator_TriCfg cfg;

			cfg.SineFrequency = GenFrequency;
			cfg.Amplitude = Amplitude;
			cfg.Offset = Offset;
			//cfg.OriginalPhase = OriginalPhase;
			cfg.OriginalPhase = 0.0f;		// Phase is unused for the moment
			cfg.Mode = 0;					// We have to generate a sine wave

			cfg.SamplingFrequency = TempSamplingFrequency;

			// Initialize with the Gen_TRI function
			Result = Gen_TRI_Init(&InputInstance[*InputInstanceIndex].pInstance, &cfg, SampleMemory);
			if (Result != FILEIO_NOERROR)
			{
				FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
				ERROR_1("Error in initializing generator: %s\n", OffsetString);
				return CMD_FILE_CANNOT_INITIALIZE_FILE;
			}
			// Initialize table
			InputInstance[*InputInstanceIndex].f_Action = &Gen_TRI_ReadSample;
			InputInstance[*InputInstanceIndex].f_Clean = &Gen_TRI_Clean;
			InputInstance[*InputInstanceIndex].f_GetPosition = &Gen_TRI_GetPosition;
			InputInstance[*InputInstanceIndex].f_GetLength = &Gen_TRI_GetFileLength;
			InputInstance[*InputInstanceIndex].f_SetLooping = &Gen_TRI_SetLoopingMode;
			*InputInstanceIndex = *InputInstanceIndex + 1;
		}
		else if (strcmp(GenName, "saw") == 0)		// The generator should generate saw tooth
		{
			Generator_TriCfg cfg;

			cfg.SineFrequency = GenFrequency;
			cfg.Amplitude = Amplitude;
			cfg.Offset = Offset;
			//cfg.OriginalPhase = OriginalPhase;
			cfg.OriginalPhase = 0.0f;		// Phase is unused for the moment
			cfg.Mode = 1;					// We have to generate a sine wave

			cfg.SamplingFrequency = TempSamplingFrequency;

			// Initialize with the Gen_TRI function
			Result = Gen_TRI_Init(&InputInstance[*InputInstanceIndex].pInstance, &cfg, SampleMemory);
			if (Result != FILEIO_NOERROR)
			{
				FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
				ERROR_1("Error in initializing generator: %s\n", OffsetString);
				return CMD_FILE_CANNOT_INITIALIZE_FILE;
			}
			// Initialize table
			InputInstance[*InputInstanceIndex].f_Action = &Gen_TRI_ReadSample;
			InputInstance[*InputInstanceIndex].f_Clean = &Gen_TRI_Clean;
			InputInstance[*InputInstanceIndex].f_GetPosition = &Gen_TRI_GetPosition;
			InputInstance[*InputInstanceIndex].f_GetLength = &Gen_TRI_GetFileLength;
			InputInstance[*InputInstanceIndex].f_SetLooping = &Gen_TRI_SetLoopingMode;
			*InputInstanceIndex = *InputInstanceIndex + 1;
		}
		else										// The input file extension is not known
		{
			ERROR_1("Sorry, the generator type: %s is not known.\n", GenName);
			return CMD_FILE_UNKNOWN_EXTENSION;
		}
	}
	
	DBG_PRINT("Out InGen\n");

	return CMD_FILE_NOERROR;
};

/************************************************************/
/* NAME       :												*/
/* DESCRIPTION:												*/
/* INPUT      :												*/
/* OUTPUT     :												*/
/************************************************************/
INT32 InWaveDev_Init(AL3101* pDspInstance, CHAR* Options, IOTable *InputInstance, INT32 *InputInstanceIndex)
{
	DBG_PRINT("In InLine\n");

	DBG_PRINT("Out InLine\n");
	
	return 0;
};


/************************************************************/
/*****        O U T P U T   F U N C T I O N S           *****/
/************************************************************/

/************************************************************/
/* NAME       :												*/
/* DESCRIPTION:												*/
/* INPUT      :												*/
/* OUTPUT     :												*/
/************************************************************/
INT32 OutFile_Init(AL3101* pDspInstance, CHAR* Options, IOTable *OutputInstance, INT32 *OutputInstanceIndex)
{
	FileIO_SndCfg cfg;
	CHAR FileName[CMD_MAX_LEN];
	CHAR TempStr[CMD_MAX_LEN];
	CHAR OffsetString[CMD_MAX_LEN];
	INT32 Offset;
	INT32 *LeftSampleMemory, *RightSampleMemory = NULL;
	INT32 RegMemSpecifier;
	INT32 TempSamplingFrequency;
	INT32 Result;
	INT32 index;

	DBG_PRINT("In OutFile\n");

	// Saving the command line for future use
	strcpy(OutputInstance[*OutputInstanceIndex].CmdLine, Options);

	// Remove the "file" keyword from the string. FileName will be overwritten with the correct string later on
	GetNextWord(FileName, Options);

	// Get the file name
	GetNextFilenameWord(FileName, Options);
	for(index = strlen(FileName); index >= 0; index--) FileName[index] = tolower(FileName[index]);

	// Extract the extension from the file name
	CHAR *pCharTemp = strrchr(FileName, '.');		// search for the '.'
	if (pCharTemp == NULL)
	{
		ERROR("Unable to determine the extension of the file.\n");
		return CMD_FILE_CANNOT_RETRIEVE_EXTENSION;
	}
	pCharTemp++;									// skip the '.' and go to the extension

	// Get the sampling frequency value
	GetNextWord(TempStr, Options);
	TempSamplingFrequency = atoi(TempStr);

	// Checking that the sampling frequency is OK with the previously initialized inputs
	if (TempSamplingFrequency != 0)						// If given sampling frequency is 0, do not take it in account
	{
		// first check the "absolute" validity of the value regarding the DSP capabilities
		if ((TempSamplingFrequency < 30000) || (TempSamplingFrequency > 50000))
		{
			WARNING("Warning: the recommended sampling frequency should be between 30kHz and 50kHz.\n");
			WARNING_1(" The sampling frequency required is %d.", TempSamplingFrequency);
			WARNING(" Simulation will go on but will perhaps not represent the real hardware operation.\n");
		}
		// then check the "relative" validity of the value regarding the inputs already set
		if (pDspInstance->SamplingFrequency == -1) pDspInstance->SamplingFrequency = TempSamplingFrequency;		// The first time, the sampling frequency used in the program is set
		else
		{
			if (pDspInstance->SamplingFrequency != TempSamplingFrequency)							// Next times, test is done
			{
				ERROR_1("The sampling frequency used in file %s\n  is different from the sampling frequency used in other inputs\n", FileName);
				ERROR(" Please check that all sampling frequencies are the same.\n");
				return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
			}
		}
	}
	else
	{
		if (TempSamplingFrequency == -1)
		{
			ERROR("The sampling frequency cannot be reused since it has not been defined yet.\n");
			ERROR("  Please define the sampling frequency with another value than 0.\n");
			ERROR("  Or use a file containing the sampling frequency in a header (like wav files) before.\n");
			return CMD_FILE_WRONG_SAMPLING_FREQUENCY;
		}
	}
	cfg.SamplingFrequency = pDspInstance->SamplingFrequency;
	

	// Get the left offset value
	// Get the reg/mem specifier or the left offset value
	GetNextWord(OffsetString, Options);
	if (OffsetString == NULL)
	{
		ERROR("Left/Mono offset not specified.\n");
		return CMD_FILE_WRONG_ARGUMENT;
	}
	GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
	switch (RegMemSpecifier)
	{
	case -1:		// unknown stuff
		ERROR("Unknown register or memory specifier. Giving up...\n");
		return CMD_FILE_WRONG_ARGUMENT;
	case 0:			// this is a number, assuming "Out" memory use
		LeftSampleMemory = pDspInstance->State.Out;
		Offset = atoi(OffsetString);
		break;
	default:		// there is a register or memory specifier
		GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &LeftSampleMemory);
		if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
		{
			GetNextWord(OffsetString, Options);
			if (OffsetString == NULL)
			{
				ERROR("Left/Mono location not specified.\n");
				return CMD_FILE_WRONG_ARGUMENT;
			}
			Offset = atoi(OffsetString);
		}
		else Offset = 0;
	}
	// Add the offset to the base address.
	LeftSampleMemory = LeftSampleMemory + Offset;


	// Get the right offset value
	// Get the reg/mem specifier or the left offset value
	GetNextWord(OffsetString, Options);
	if (OffsetString == NULL)
	{
		ERROR("Left/Mono offset not specified.\n");
		return CMD_FILE_WRONG_ARGUMENT;
	}
	else if (strcmp(OffsetString, "") == 0)
	{
		Offset = -1;
		RightSampleMemory = NULL;
		cfg.NumberOfChannels = 1;		// The WAV file is monophonic
	}
	else
	{
		GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
		switch (RegMemSpecifier)
		{
		case -1:		// unknown stuff
			ERROR("Unknown register or memory specifier. Giving up...\n");
			return CMD_FILE_WRONG_ARGUMENT;
		case 0:			// this is a number, assuming "Out" memory use
			RightSampleMemory = pDspInstance->State.Out;
			Offset = atoi(OffsetString);
			break;
		default:		// there is a register or memory specifier
			GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &RightSampleMemory);
			if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
			{
				GetNextWord(OffsetString, Options);
				if (OffsetString == NULL)
				{
					ERROR("Right location not specified.\n");
					return CMD_FILE_WRONG_ARGUMENT;
				}
				Offset = atoi(OffsetString);
			}
			else Offset = 0;
		}
		// Add the offset to the base address.
		RightSampleMemory = RightSampleMemory + Offset;
		cfg.NumberOfChannels = 2;		// The WAV file is stereophonic
	}

	
	// making the file type specific things
	if (strcmp(pCharTemp, "wav") == 0)			// The input file is a .wav file
	{
		// Initialize the rest of the cfg structure
		cfg.NumberOfBytesInSample = 2;					// The WAV file will contain 16-bits samples
		// Initialize with the FILEIO wav function
		Result = FileIO_WAV_Init(&OutputInstance[*OutputInstanceIndex].pInstance, FileName, 1, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		OutputInstance[*OutputInstanceIndex].f_Action = &FileIO_WAV_WriteSample;
		OutputInstance[*OutputInstanceIndex].f_Clean = &FileIO_WAV_Clean;
		OutputInstance[*OutputInstanceIndex].f_GetPosition = &FileIO_WAV_GetPosition;
		OutputInstance[*OutputInstanceIndex].f_GetLength = &FileIO_WAV_GetFileLength;
		*OutputInstanceIndex = *OutputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "pcm") == 0)		// The input file is a .pcm file
	{
		// Initialize the rest of the cfg structure
		cfg.NumberOfBytesInSample = 2;					// The WAV file will contain 16-bits samples
		// Initialize with the FILEIO wav function
		Result = FileIO_PCM_Init(&OutputInstance[*OutputInstanceIndex].pInstance, FileName, 1, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		OutputInstance[*OutputInstanceIndex].f_Action = &FileIO_PCM_WriteSample;
		OutputInstance[*OutputInstanceIndex].f_Clean = &FileIO_PCM_Clean;
		OutputInstance[*OutputInstanceIndex].f_GetPosition = &FileIO_PCM_GetPosition;
		OutputInstance[*OutputInstanceIndex].f_GetLength = &FileIO_PCM_GetFileLength;
		*OutputInstanceIndex = *OutputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "dat") == 0)		// The input file is a .dat file
	{
		// Initialize the rest of the cfg structure
		cfg.NumberOfBytesInSample = 2;					// This value is unused, let's set it to 2
		// Reset the number of channels
		if (cfg.NumberOfChannels != 1)				// explaining that left only is handled
		{
			WARNING("Dat files only support mono.\nRight channel will be ignored.\n");
		}
		cfg.NumberOfChannels = 1;						// The .dat files only support mono
		// Initialize with the FILEIO dat function
		Result = FileIO_DAT_Init(&OutputInstance[*OutputInstanceIndex].pInstance, FileName, 1, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		OutputInstance[*OutputInstanceIndex].f_Action = &FileIO_DAT_WriteSample;
		OutputInstance[*OutputInstanceIndex].f_Clean = &FileIO_DAT_Clean;
		OutputInstance[*OutputInstanceIndex].f_GetPosition = &FileIO_DAT_GetPosition;
		OutputInstance[*OutputInstanceIndex].f_GetLength = &FileIO_DAT_GetFileLength;
		*OutputInstanceIndex = *OutputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "dec") == 0)		// The input file is a .dec file
	{
		// Initialize the rest of the cfg structure
		cfg.NumberOfBytesInSample = 2;					// This value is unused, let's set it to 2
		// Reset the number of channels
		if (cfg.NumberOfChannels != 1)				// explaining that left only is handled
		{
			WARNING("Int files only support mono.\nRight channel will be ignored.\n");
		}
		cfg.NumberOfChannels = 1;						// The .dat files only support mono
		// Initialize with the FILEIO dec function
		Result = FileIO_DEC_Init(&OutputInstance[*OutputInstanceIndex].pInstance, FileName, 1, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		OutputInstance[*OutputInstanceIndex].f_Action = &FileIO_DEC_WriteSample;
		OutputInstance[*OutputInstanceIndex].f_Clean = &FileIO_DEC_Clean;
		OutputInstance[*OutputInstanceIndex].f_GetPosition = &FileIO_DEC_GetPosition;
		OutputInstance[*OutputInstanceIndex].f_GetLength = &FileIO_DEC_GetFileLength;
		*OutputInstanceIndex = *OutputInstanceIndex + 1;
	}
	else if (strcmp(pCharTemp, "hex") == 0)		// The input file is a .hex file
	{
		// Initialize the rest of the cfg structure
		cfg.NumberOfBytesInSample = 2;					// This value is unused, let's set it to 2
		// Reset the number of channels
		if (cfg.NumberOfChannels != 1)				// explaining that left only is handled
		{
			WARNING("Hex files only support mono.\nRight channel will be ignored.\n");
		}
		cfg.NumberOfChannels = 1;						// The .dat files only support mono
		// Initialize with the FILEIO hex function
		Result = FileIO_HEX_Init(&OutputInstance[*OutputInstanceIndex].pInstance, FileName, 1, &cfg, LeftSampleMemory, RightSampleMemory);
		if (Result != FILEIO_NOERROR)
		{
			FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing input file: %s\n", OffsetString);
			return CMD_FILE_CANNOT_INITIALIZE_FILE;
		}
		// Initialize table
		OutputInstance[*OutputInstanceIndex].f_Action = &FileIO_HEX_WriteSample;
		OutputInstance[*OutputInstanceIndex].f_Clean = &FileIO_HEX_Clean;
		OutputInstance[*OutputInstanceIndex].f_GetPosition = &FileIO_HEX_GetPosition;
		OutputInstance[*OutputInstanceIndex].f_GetLength = &FileIO_HEX_GetFileLength;
		*OutputInstanceIndex = *OutputInstanceIndex + 1;
	}
	else										// The input file extension is not known
	{
		ERROR_1("Sorry, the file extension: %s is not known.\n", pCharTemp);
		return CMD_FILE_UNKNOWN_EXTENSION;
	}
  
	DBG_PRINT("Out OutFile\n");
	
	return CMD_FILE_NOERROR;
};

/************************************************************/
/* NAME       :												*/
/* DESCRIPTION:												*/
/* INPUT      :												*/
/* OUTPUT     :												*/
/************************************************************/
INT32 OutWaveDev_Init(AL3101* pDspInstance, CHAR* Options, IOTable *OutputInstance, INT32 *OutputInstanceIndex)
{
	DBG_PRINT("In OutWaveDev\n");
	
	DBG_PRINT("Out OutWaveDev\n");
	
	return 0;
};


/************************************************************/
/*****          TOOLS   F U N C T I O N S               *****/
/************************************************************/

/************************************************************/
/* NAME       :												*/
/* DESCRIPTION:												*/
/* INPUT      :												*/
/* OUTPUT     :												*/
/************************************************************/
INT32 Tool_Init(AL3101* pDspInstance, CHAR* Options)
{
	FileIO_SndCfg cfg;
	CHAR ToolName[CMD_MAX_LEN];
	CHAR TempStr[CMD_MAX_LEN];
	INT32 TempSamplingFrequency;
	INT32 Result;
	INT32 index;

	DBG_PRINT("In Tool_Init\n");

	// Saving the command line for future use
	strcpy(pDspInstance->Tools[pDspInstance->NumberOfToolInstances].CmdLine, Options);

	// Get the tool name
	GetNextWord(ToolName, Options);
	for(index = strlen(ToolName); index >= 0; index--) ToolName[index] = tolower(ToolName[index]);

	// Get the sampling frequency value
	GetNextWord(TempStr, Options);
	TempSamplingFrequency = atoi(TempStr);

	// Checking that the sampling frequency is OK with the previously initialized inputs
	if (TempSamplingFrequency != 0)						// If given sampling frequency is 0, do not take it in account
	{
		// first check the "absolute" validity of the value regarding the DSP capabilities
		if ((TempSamplingFrequency < 30000) || (TempSamplingFrequency > 50000))
		{
			WARNING("Warning: the recommended sampling frequency should be between 30kHz and 50kHz.\n");
			WARNING_1(" The sampling frequency required is %d.", TempSamplingFrequency);
			WARNING(" Simulation will go on but will perhaps not represent the real hardware operation.\n");
		}
		// then check the "relative" validity of the value regarding the inputs already set
		if (pDspInstance->SamplingFrequency == -1) pDspInstance->SamplingFrequency = TempSamplingFrequency;		// The first time, the sampling frequency used in the program is set
		else
		{
			if (pDspInstance->SamplingFrequency != TempSamplingFrequency)							// Next times, test is done
			{
				ERROR_1("The sampling frequency used in tool %s\n  is different from the sampling frequency used in other inputs\n", ToolName);
				ERROR(" Please check that all sampling frequencies are the same.\n");
				return CMD_TOOL_WRONG_SAMPLING_FREQUENCY;
			}
		}
	}
	else
	{
		if (TempSamplingFrequency == -1)
		{
			ERROR("The sampling frequency cannot be reused since it has not been defined yet.\n");
			ERROR("  Please define the sampling frequency with another value than 0.\n");
			ERROR("  Or use a file containing the sampling frequency in a header (like wav files) before.\n");
			return CMD_TOOL_WRONG_SAMPLING_FREQUENCY;
		}
	}
	cfg.SamplingFrequency = pDspInstance->SamplingFrequency;

	// making the tool specific things
	if (strcmp(ToolName, "ampsweep") == 0)			// The tool is the amplitude sweep
	{
		Tools_AmplitudeSweep_LevelParameters LevelParams;
		Tools_AmplitudeSweep_TimingParameters TimingParams;
		CHAR AmplitudeFileName[CMD_MAX_LEN];
		CHAR GainFileName[CMD_MAX_LEN];
		CHAR OffsetString[CMD_MAX_LEN];
		INT32 Offset;
		INT32 *InputSampleMemory, *OutputSampleMemory;
		INT32 RegMemSpecifier;

		// Get the Sine operation frequency value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Sine operating frequency not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		TimingParams.SineFrequency = (float)atof(OffsetString);
		TimingParams.SamplingFrequency = (float)cfg.SamplingFrequency;
	
		// Get the transient time value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Transient time not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		TimingParams.TransientTime = (float)atof(OffsetString);
	
		// Get the measurement time value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Measurement time not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		TimingParams.MeasurementTime = (float)atof(OffsetString);
	
		// Get the Initial Level value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Initial level not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		LevelParams.InitialLevel = (float)atof(OffsetString);
	
		// Get the Final Level value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Final level not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		LevelParams.FinalLevel = (float)atof(OffsetString);
	
		// Get the Level Increment value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Level increment not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		LevelParams.LevelIncrement = (float)atof(OffsetString);
	
		// Get the input offset value
		// Get the reg/mem specifier or the left offset value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Input position not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
		switch (RegMemSpecifier)
		{
		case -1:		// unknown stuff
			ERROR("Unknown register or memory specifier. Giving up...\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		case 0:			// this is a number, assuming "In" memory use
			InputSampleMemory = pDspInstance->State.In;
			Offset = atoi(OffsetString);
			break;
		default:		// there is a register or memory specifier
			GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &InputSampleMemory);
			if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
			{
				GetNextWord(OffsetString, Options);
				if (OffsetString == NULL)
				{
					ERROR("Input position not specified.\n");
					return CMD_TOOL_WRONG_ARGUMENT;
				}
				Offset = atoi(OffsetString);
			}
			else Offset = 0;
		}
		// Add the offset to the base address.
		InputSampleMemory = InputSampleMemory + Offset;
	
		// Get the output offset value
		// Get the reg/mem specifier or the left offset value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Output position not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
		switch (RegMemSpecifier)
		{
		case -1:		// unknown stuff
			ERROR("Unknown register or memory specifier. Giving up...\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		case 0:			// this is a number, assuming "Out" memory use
			OutputSampleMemory = pDspInstance->State.Out;
			Offset = atoi(OffsetString);
			break;
		default:		// there is a register or memory specifier
			GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &OutputSampleMemory);
			if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
			{
				GetNextWord(OffsetString, Options);
				if (OffsetString == NULL)
				{
					ERROR("Output position not specified.\n");
					return CMD_TOOL_WRONG_ARGUMENT;
				}
				Offset = atoi(OffsetString);
			}
			else Offset = 0;
		}
		// Add the offset to the base address.
		OutputSampleMemory = OutputSampleMemory + Offset;

		// Get the amplitude file name
		GetNextFilenameWord(AmplitudeFileName, Options);

		// Get the gain file name
		GetNextFilenameWord(GainFileName, Options);
	
		// Initialize with the AmpSweep init function
		Result = Tool_AmplitudeSweep_Init(&pDspInstance->Tools[pDspInstance->NumberOfToolInstances].pInstance, NULL, &LevelParams, &TimingParams, InputSampleMemory, OutputSampleMemory, AmplitudeFileName, GainFileName);
		if (Result != 0)
		{
			//FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing tool: %s\n", ToolName);
			return CMD_TOOL_CANNOT_INITIALIZE_TOOL;
		}
		// Initialize table
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_Read = &Tool_AmplitudeSweep_ReadSample;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_Write = &Tool_AmplitudeSweep_WriteSample;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_Clean = &Tool_AmplitudeSweep_Clean;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_GetPosition = &Tool_AmplitudeSweep_GetPosition;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_GetLength = &Tool_AmplitudeSweep_GetFileLength;
		pDspInstance->NumberOfToolInstances++;
	}
	else if (strcmp(ToolName, "freqresp") == 0)			// The tool is the frequency response
	{
		Tools_FreqResp_Parameters Params;
		CHAR AmplitudeFileName[CMD_MAX_LEN];
		CHAR PhaseFileName[CMD_MAX_LEN];
		CHAR OffsetString[CMD_MAX_LEN];
		INT32 Offset;
		INT32 *InputSampleMemory, *OutputSampleMemory;
		INT32 RegMemSpecifier;

		Params.SamplingFrequency = (float)cfg.SamplingFrequency;

		// Get the BlockSize value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Block Size not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		Params.BlockSize = atoi(OffsetString);
	
		// Get the NSkipBlocks value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Number of blocks to skip not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		Params.NSkipBlocks = atoi(OffsetString);
	
		// Get the Noise Level value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Noise level not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		Params.NoiseLevel = (float)atof(OffsetString);
	
		// Get the input offset value
		// Get the reg/mem specifier or the left offset value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Input position not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
		switch (RegMemSpecifier)
		{
		case -1:		// unknown stuff
			ERROR("Unknown register or memory specifier. Giving up...\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		case 0:			// this is a number, assuming "In" memory use
			InputSampleMemory = pDspInstance->State.In;
			Offset = atoi(OffsetString);
			break;
		default:		// there is a register or memory specifier
			GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &InputSampleMemory);
			if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
			{
				GetNextWord(OffsetString, Options);
				if (OffsetString == NULL)
				{
					ERROR("Input position not specified.\n");
					return CMD_TOOL_WRONG_ARGUMENT;
				}
				Offset = atoi(OffsetString);
			}
			else Offset = 0;
		}
		// Add the offset to the base address.
		InputSampleMemory = InputSampleMemory + Offset;
	
		// Get the output offset value
		// Get the reg/mem specifier or the left offset value
		GetNextWord(OffsetString, Options);
		if (OffsetString == NULL)
		{
			ERROR("Output position not specified.\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		}
		GetRegMemSpecifier(OffsetString, &RegMemSpecifier);
		switch (RegMemSpecifier)
		{
		case -1:		// unknown stuff
			ERROR("Unknown register or memory specifier. Giving up...\n");
			return CMD_TOOL_WRONG_ARGUMENT;
		case 0:			// this is a number, assuming "Out" memory use
			OutputSampleMemory = pDspInstance->State.Out;
			Offset = atoi(OffsetString);
			break;
		default:		// there is a register or memory specifier
			GetAddressStringFromSpecifier(RegMemSpecifier, TempStr, &OutputSampleMemory);
			if (RegMemSpecifier > REGMEM_SPECIFIER_LIMIT)
			{
				GetNextWord(OffsetString, Options);
				if (OffsetString == NULL)
				{
					ERROR("Output position not specified.\n");
					return CMD_TOOL_WRONG_ARGUMENT;
				}
				Offset = atoi(OffsetString);
			}
			else Offset = 0;
		}
		// Add the offset to the base address.
		OutputSampleMemory = OutputSampleMemory + Offset;

		// Get the amplitude file name
		GetNextFilenameWord(AmplitudeFileName, Options);

		// Get the phase file name
		GetNextFilenameWord(PhaseFileName, Options);
	
		// Initialize with the FreqResp init function
		Result = Tool_FreqResp_Init(&pDspInstance->Tools[pDspInstance->NumberOfToolInstances].pInstance, NULL, &Params, InputSampleMemory, OutputSampleMemory, AmplitudeFileName, PhaseFileName);
		if (Result != 0)
		{
			//FileIO_GetLastErrorString(OffsetString, Result, CMD_MAX_LEN-1);
			ERROR_1("Error in initializing tool: %s\n", ToolName);
			return CMD_TOOL_CANNOT_INITIALIZE_TOOL;
		}
		// Initialize table
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_Read = &Tool_FreqResp_ReadSample;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_Write = &Tool_FreqResp_WriteSample;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_Clean = &Tool_FreqResp_Clean;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_GetPosition = &Tool_FreqResp_GetPosition;
		pDspInstance->Tools[pDspInstance->NumberOfToolInstances].f_GetLength = &Tool_FreqResp_GetFileLength;
		pDspInstance->NumberOfToolInstances++;
	}
	/*else if (strcmp(ToolName, "...") == 0)		// The tool is ...
	{
	}*/
	else										// The input file extension is not known
	{
		ERROR_1("Sorry, the tool name : %s is not known.\n", ToolName);
		return CMD_TOOL_UNKNOWN;
	}
  
	DBG_PRINT("Out Tool_Init\n");
	
	return CMD_TOOL_NOERROR;
}


