// Tool_AmplitudeSweep.cpp: implementation of the AmplitudeSweep tool functions.
//
/////////////////////////////////////////////////////////////////////////////////

#include "../MessageMacros.h"		// defines the macros for displaying macros
#include "Tool_AmplitudeSweep.h"
#include "Tool_AmplitudeSweep_Internal.h"

#include <stdlib.h>		// for the malloc function
#include <string.h>		// for the string routines
#include <math.h>		// for the log function


//////////////////////////////////////////////////////////////////////
// init & clean functions
//////////////////////////////////////////////////////////////////////
INT32 Tool_AmplitudeSweep_Init(void **pInstance, Tool_AmplitudeSweep_Results *results, Tools_AmplitudeSweep_LevelParameters *LevelParams, Tools_AmplitudeSweep_TimingParameters *TimingParams, INT32 *InputMemory, INT32 *OutputMemory, CHAR *AmplitudeFileName, CHAR *GainFileName)
{
	Tool_AmplitudeSweep_Struct *pInst;
	INT32 ResultSize, ReturnError;

	// test the level parameters and get the size of results as side effect
	ReturnError = Tool_AmplitudeSweep_GetResultSize(LevelParams, &ResultSize);
	if (ReturnError != 0) return -2;	// WRONG_ARGUMENT
	// test the timing parameters
	if ((TimingParams->TransientTime < 0) || (TimingParams->MeasurementTime < 0)) return -2; //WRONG_ARGUMENT
	if (TimingParams->SineFrequency > (TimingParams->SamplingFrequency / 2)) return -2; // WRONG_ARGUMENT


	// Allocate the space for the instance structure
	pInst = new Tool_AmplitudeSweep_Struct;
	if (pInst == NULL) return -1; //TOOL_CANNOT_ALLOCATE_INSTANCE
	*pInstance = (void *)pInst;


	// Initialize result related variables
	if (results == NULL)		// command-line version. Allocation has to be made, filenames to be kept
	{
		pInst->Results = new Tool_AmplitudeSweep_Results[ResultSize];
		if (pInst->Results == NULL) return -1;	//TOOL_CANNOT_ALLOCATE_INSTANCE
		pInst->ResultsAllocatedByTool = 1;
	}
	else						// GUI version. Allocation is done at application level and filename not given
	{
		pInst->Results = results;
		pInst->ResultsAllocatedByTool = 0;
	}
	pInst->ResultsIndex = 0;
	pInst->ResultSize = ResultSize;

	// Initialize the filenames
	pInst->AmplitudeFileName = NULL;
	pInst->GainFileName = NULL;
	if (AmplitudeFileName != NULL)
	{
		strcpy(pInst->AmplitudeFileNameString, AmplitudeFileName);
		pInst->AmplitudeFileName = pInst->AmplitudeFileNameString;
	}
	if (GainFileName != NULL)
	{
		strcpy(pInst->GainFileNameString, GainFileName);
		pInst->GainFileName = pInst->GainFileNameString;
	}


	// Initialize level related internal variables
	pInst->InitialLevel = LevelParams->InitialLevel;
	pInst->FinalLevel = LevelParams->FinalLevel;
	pInst->LevelIncrement = LevelParams->LevelIncrement;
	pInst->InputAmplitudeMax = 0;
	pInst->OutputAmplitudeMax = 0;
	
	// Initialize timing related internal variables
	pInst->TransientCount = (INT32)(TimingParams->TransientTime * TimingParams->SamplingFrequency);
	pInst->TotalCount = (INT32)((TimingParams->TransientTime + TimingParams->MeasurementTime) * TimingParams->SamplingFrequency);
	pInst->SineFrequency = TimingParams->SineFrequency;
	pInst->SamplingFrequency = TimingParams->SamplingFrequency;

	if (pInst->TransientCount == 0)
	{
		WARNING("Tool AmplitudeSweep Warning: Transient time set to a stupidly small value.\n");
		pInst->TransientCount += 1;
	}
	if (pInst->TotalCount == 0)
	{
		WARNING("Tool AmplitudeSweep Warning: Measurement time set to a stupidly small value.\n");
		pInst->TotalCount += 2;
	}

	// Initialize other internal variables
	pInst->CurrentCount = 0;
	pInst->OverallCount = 0;
	pInst->CurrentLevel = LevelParams->InitialLevel;
	pInst->CurrentLevelLinear = (float)pow(10, pInst->CurrentLevel / 20.0f);


	// Initialize memory related variables
	pInst->InputMemory = InputMemory;
	pInst->OutputMemory = OutputMemory;

	// Initialize the Sine generator
	pInst->SinGenerator.Init(TimingParams->SamplingFrequency, pInst->SineFrequency, 1.0f, 0.0f);

	return 0;	// NO_ERROR
}

INT32 Tool_AmplitudeSweep_Clean(void *pInstance)
{
	Tool_AmplitudeSweep_Struct *pInst = (Tool_AmplitudeSweep_Struct *)pInstance;

	if (pInst->AmplitudeFileName != NULL)
		if (Tool_AmplitudeSweep_WriteAmplitudeResults(pInst->Results, pInst->ResultSize, pInst->AmplitudeFileName) != 0)
			ERROR("Tool Amplitude Sweep: Cannot write amplitude result report file.\n");
	if (pInst->GainFileName != NULL)
		if (Tool_AmplitudeSweep_WriteGainResults(pInst->Results, pInst->ResultSize, pInst->GainFileName) != 0)
			ERROR("Tool Amplitude Sweep: Cannot write amplitude result report file.\n");

	if (pInst->ResultsAllocatedByTool == 1)
		delete[] pInst->Results;

	// freeing the instance structure
	if (pInst != NULL) delete pInst;

	return 0;	// NO_ERROR
}


//////////////////////////////////////////////////////////////////////
// "processing" functions
//////////////////////////////////////////////////////////////////////
INT32 Tool_AmplitudeSweep_ReadSample(void *pInstance)
{
	float SinValue;
	INT32 NewSample;

	Tool_AmplitudeSweep_Struct *pInst = (Tool_AmplitudeSweep_Struct *)pInstance;

	// Generate new sample
	pInst->SinGenerator.Process(&SinValue);						// Generate the new sine value
	SinValue = SinValue * pInst->CurrentLevelLinear;			// Apply the current amplitude
	NewSample = (INT32)(SinValue * 16777216.0f);				// transform to S3.24
	*(pInst->InputMemory) = NewSample;							// copy sine value to memory

	// Update InputAmplitudeMax if needed (i.e. if TransientTime over)
	if (pInst->CurrentCount > pInst->TransientCount)			// need to check max amplitude
		if (abs(NewSample) > pInst->InputAmplitudeMax)			// check for maximum
			pInst->InputAmplitudeMax = abs(NewSample);			// update maximum if necessary

	return 0;
}

INT32 Tool_AmplitudeSweep_WriteSample(void *pInstance)
{
	INT32 NewSample;
	float Gain;
	float temp;

	Tool_AmplitudeSweep_Struct *pInst = (Tool_AmplitudeSweep_Struct *)pInstance;

	// Take the new sample
	NewSample = *(pInst->OutputMemory);

	// Update OutputAmplitudeMax if needed (i.e. if TransientTime over)
	if (pInst->CurrentCount > pInst->TransientCount)
	{
		if (abs(NewSample) > pInst->OutputAmplitudeMax) pInst->OutputAmplitudeMax = abs(NewSample);
	}

	// Update counters
	pInst->CurrentCount += 1;
	pInst->OverallCount += 1;

	// Test if TotalTime limit has been reached
	if (pInst->CurrentCount > pInst->TotalCount)
	{	// if so, Calculate Gain
		Gain = (float)pInst->OutputAmplitudeMax / (float)pInst->InputAmplitudeMax;

		// Store amplitudes and gains
		temp = (float)pInst->InputAmplitudeMax / 16777216.0f;
		temp = 20.0f * (float)log10(temp);
		(pInst->Results[pInst->ResultsIndex]).InputAmplitude = temp;
		temp = (float)pInst->OutputAmplitudeMax / 16777216.0f;
		temp = 20.0f * (float)log10(temp);
		(pInst->Results[pInst->ResultsIndex]).OutputAmplitude = temp;
		Gain = 20.0f * (float)log10(Gain);
		(pInst->Results[pInst->ResultsIndex]).Gain = Gain;

		// Adapt the parameters for next samples
		pInst->ResultsIndex += 1;
		pInst->CurrentLevel += pInst->LevelIncrement;
		pInst->CurrentLevelLinear = (float)pow(10, pInst->CurrentLevel / 20.0f);
		pInst->CurrentCount = 0;

		// Check limits for end of simulation
		if (pInst->CurrentLevel > pInst->FinalLevel)
		{	// if so, clean if needed and return END_OF_SIMULATION
			return -1;	// END_OF_SIMULATION
		}
	}
	return 0;
}





//////////////////////////////////////////////////////////////////////
// file-like function(s)
//////////////////////////////////////////////////////////////////////
INT32 Tool_AmplitudeSweep_GetPosition(void *pInstance, INT32 *CurrentPosition)
{
	Tool_AmplitudeSweep_Struct *pInst = (Tool_AmplitudeSweep_Struct *)pInstance;

	*CurrentPosition = pInst->OverallCount;

	return 0;	//TOOL_NOERROR
}

INT32 Tool_AmplitudeSweep_GetFileLength(void *pInstance, INT32 *FileSize)
{
	INT32 index = 1;
	float CurrentLevel;

	Tool_AmplitudeSweep_Struct *pInst = (Tool_AmplitudeSweep_Struct *)pInstance;

	// Get the number of increment to apply
	CurrentLevel = pInst->InitialLevel;
	while (CurrentLevel < pInst->FinalLevel)
	{
		CurrentLevel += pInst->LevelIncrement;
		index++;
	}
	// At this stage, index contains the number of step increments.
	// TotalCount is the total number of steps required for a step increment.
	*FileSize = pInst->TotalCount * index;
	
	return 0; //TOOL_NOERROR;
}



/*//////////////////////////////////////////////////////////////////////
// control function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_PCM_SetLoopingMode(void *pInstance, INT32 Looping)
{
	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;
	pInst->Loop = Looping;

	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// "text" function(s)
//////////////////////////////////////////////////////////////////////
void FileIO_PCM_GetLastErrorString(CHAR *str, INT32 LastErrorCode, INT32 MaxNumberOfCharacters)
{
	if ((LastErrorCode & FILEIO_BASE_ERR) == FILEIO_BASE_ERR)		// check if the error is a FileIO error
	{
		if ((LastErrorCode & FILEIO_GEN_BASE_ERR) == FILEIO_GEN_BASE_ERR)
		{
			FileIO_GetLastErrorString(str, LastErrorCode, MaxNumberOfCharacters);
			return;
		}
		else if ((LastErrorCode & FILEIO_PCM_BASE_ERR) == FILEIO_PCM_BASE_ERR)
		{
			// do nothing here, wait for the switch below
		}
		else
		{
			strncpy(str, "Error type not known", MaxNumberOfCharacters - 1);
		}
	}
	else															// this error is not a FileIO error
	{
		strncpy(str, "Not a FileIO error", MaxNumberOfCharacters - 1);
		return;
	}


	switch(LastErrorCode)
	{
		case FILEIO_PCM_NOERROR:
			strncpy(str, "PCM: No error occurred", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "PCM: Error not known", MaxNumberOfCharacters - 1);
			break;
	}
	return;
}



void FileIO_PCM_GetHelp(CHAR *str, INT32 LineNumber, INT32 MaxNumberOfCharacters)
{
	switch(LineNumber)
	{
		case 0:
			strncpy(str, "PCM files are read by these functions.", MaxNumberOfCharacters - 1);
			break;
		case 1:
			strncpy(str, "PCM files are raw data encoded in Little Endian format i.e. LSB,MSB", MaxNumberOfCharacters - 1);
			break;
		case 2:
			strncpy(str, "There's no support for 8 bits files and for more than 4GB files.", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "PCM file functions", MaxNumberOfCharacters - 1);
	}
}
*/





//////////////////////////////////////////////////////////////////////
// Helper function(s)
//////////////////////////////////////////////////////////////////////
INT32 Tool_AmplitudeSweep_GetResultSize(Tools_AmplitudeSweep_LevelParameters *LevelParams, INT32 *ResultSize)
{
	INT32 index = 1;
	float CurrentLevel = LevelParams->InitialLevel;

	// Test validity of parameters
	if ((LevelParams->FinalLevel <= LevelParams->InitialLevel) || (LevelParams->LevelIncrement <= 0))
	{
		*ResultSize = -1;
		return -1; //WRONG_ARGUMENT
	}
	
	while (CurrentLevel < LevelParams->FinalLevel)
	{
		CurrentLevel += LevelParams->LevelIncrement;
		index++;
	}

	*ResultSize = index;
	
	return 0;	// NO_ERROR
}


INT32 Tool_AmplitudeSweep_WriteAmplitudeResults(Tool_AmplitudeSweep_Results *results, INT32 ResultSize, CHAR *AmplitudeFileName)
{
	FILE *fAmplitude;
	INT32 index;

	// Open the amplitude file
	fAmplitude = fopen(AmplitudeFileName, "w+t");
	if (fAmplitude == NULL) return -1;

	for(index = 0; index < ResultSize; index++)
	{
		fprintf(fAmplitude, "%g\t%g\n", results[index].InputAmplitude, results[index].OutputAmplitude);
	}

	fclose(fAmplitude);

	return 0;// NO_ERROR;
}

INT32 Tool_AmplitudeSweep_WriteGainResults(Tool_AmplitudeSweep_Results *results, INT32 ResultSize, CHAR *GainFileName)
{
	FILE *fGain;
	INT32 index;

	// Open the amplitude file
	fGain = fopen(GainFileName, "w+t");
	if (fGain == NULL) return -1;

	for(index = 0; index < ResultSize; index++)
	{
		fprintf(fGain, "%g\t%g\n", results[index].InputAmplitude, results[index].Gain);
	}

	fclose(fGain);

	return 0;// NO_ERROR;
}
