// Tool_FrequencyResponse.cpp: implementation of the FrequencyResponse tool functions.
//
///////////////////////////////////////////////////////////////////////////////////////

#include "../MessageMacros.h"		// defines the macros for displaying macros
#include "Tool_FrequencyResponse.h"
#include "Tool_FrequencyResponse_Internal.h"
#include "FFTWindowing.h"

#include <stdlib.h>		// for the malloc function
#include <string.h>		// for the string routines
#include <math.h>		// for the log function


//////////////////////////////////////////////////////////////////////
// init & clean functions
//////////////////////////////////////////////////////////////////////
INT32 Tool_FreqResp_Init(void **pInstance, Tool_FreqResp_Results *results, Tools_FreqResp_Parameters *Params, INT32 *InputMemory, INT32 *OutputMemory, CHAR *AmplitudeFileName, CHAR *PhaseFileName)
{
	Tool_FreqResp_Struct *pInst;
	INT32 ResultSize, ReturnError;

	// test the parameters and get the size of results as side effect
	ReturnError = Tool_FreqResp_GetResultSize(Params, &ResultSize);
	if (ReturnError != 0) return -2;	// WRONG_ARGUMENT
	// test the timing parameters
	if (Params->NSkipBlocks < 1)
	{
		ERROR_1("NSkipBlocks too small: %i, value should be >= 1", Params->NSkipBlocks);
		return -2; //WRONG_ARGUMENT
	}


	// Allocate the space for the instance structure
	pInst = new Tool_FreqResp_Struct;
	if (pInst == NULL) return -1; //TOOL_CANNOT_ALLOCATE_INSTANCE
	*pInstance = (void *)pInst;


	// Initialize result related variables
	if (results == NULL)		// command-line version. Allocation has to be made, filenames to be kept
	{
		pInst->Results = new Tool_FreqResp_Results[ResultSize + 1];
		if (pInst->Results == NULL)
		{
			delete pInst;
			return -1;	//TOOL_CANNOT_ALLOCATE_INSTANCE
		}
		pInst->ResultsAllocatedByTool = 1;
	}
	else						// GUI version. Allocation is done at application level and filename not given
	{
		pInst->Results = results;
		pInst->ResultsAllocatedByTool = 0;
	}


	// Initialize timing related internal variables
	if (Params->BlockSize < 4)
	{
		WARNING_1("Tool FreqResp Warning: Block size set to a stupidly small value: %i.\n", Params->BlockSize);
	}
	if (Params->NSkipBlocks < 1)
	{
		WARNING_1("Tool FreqResp Warning: Number of block to skip set to a stupidly small value: %i.\n", Params->NSkipBlocks);
		Params->NSkipBlocks = Params->NSkipBlocks + 1;
		WARNING_1("Tool FreqResp Warning: Number of block to skip set to %i.\n", Params->NSkipBlocks);
	}
	pInst->BlockSize = Params->BlockSize;
	pInst->NSkipBlocks = Params->NSkipBlocks;

	pInst->SkippingCounter = (INT32)(pInst->NSkipBlocks * pInst->BlockSize);
	pInst->ProcessingCounter = pInst->BlockSize;
	pInst->OverallCounter = 0;
	pInst->MeasuringMode = 0;
	pInst->SamplingFrequency = Params->SamplingFrequency;

	// Allocate memory for the input and output buffers when measuring
	pInst->InputBuffer = new float[pInst->BlockSize];
	if (pInst->InputBuffer == NULL)
	{
		delete pInst->Results;
		delete pInst;
		return -1;
	}
	pInst->OutputBuffer = new float[pInst->BlockSize];
	if (pInst->OutputBuffer == NULL)
	{
		delete pInst->OutputBuffer;
		delete pInst->Results;
		delete pInst;
		return -1;
	}

	// Initialize the filenames
	pInst->AmplitudeFileName = NULL;
	pInst->PhaseFileName = NULL;
	if (AmplitudeFileName != NULL)
	{
		strcpy(pInst->AmplitudeFileNameString, AmplitudeFileName);
		pInst->AmplitudeFileName = pInst->AmplitudeFileNameString;
	}
	if (PhaseFileName != NULL)
	{
		strcpy(pInst->PhaseFileNameString, PhaseFileName);
		pInst->PhaseFileName = pInst->PhaseFileNameString;
	}

	// Initialize memory related variables
	pInst->InputMemory = InputMemory;
	pInst->OutputMemory = OutputMemory;

	// Initialize the Noise related parameters
	if (Params->NoiseLevel >= 18.0)
		WARNING("Tool FreqResp Warning: The Noise level is set higher than 18dB. A strange behaviour might occur.\n");
	pInst->NoiseLevel = Params->NoiseLevel;
	pInst->NoiseLevelLinear = (float)pow(10, (Params->NoiseLevel / 20));

	return 0;	// NO_ERROR
}

INT32 Tool_FreqResp_Clean(void *pInstance)
{
	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;

	if (Tool_FreqResp_ComputeResults(pInstance) != 0)
	{
		ERROR("Unable to compute the FFT on the signals.\n");
	}

	if (pInst->AmplitudeFileName != NULL)
		if (Tool_FreqResp_WriteAmplitudeResults(pInst->Results, pInst->BlockSize / 2, pInst->AmplitudeFileName) != 0)
			ERROR("Tool FreqResp: Cannot write amplitude result report file.\n");
	if (pInst->PhaseFileName != NULL)
		if (Tool_FreqResp_WritePhaseResults(pInst->Results, pInst->BlockSize / 2, pInst->PhaseFileName) != 0)
			ERROR("Tool FreqResp: Cannot write phase result report file.\n");

	delete pInst->InputBuffer;
	delete pInst->OutputBuffer;

	if (pInst->ResultsAllocatedByTool == 1)
		delete[] pInst->Results;

	// freeing the instance structure
	if (pInst != NULL) delete pInst;

	return 0;	// NO_ERROR
}


//////////////////////////////////////////////////////////////////////
// "processing" functions
//////////////////////////////////////////////////////////////////////
INT32 Tool_FreqResp_ReadSample(void *pInstance)
{
	float NoiseValue;
	INT32 NewSample;

	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;

	// Generate new sample
	pInst->NoiseGenerator.Process(&NoiseValue);					// Generate the new noise value
	NoiseValue = NoiseValue * pInst->NoiseLevelLinear;			// Apply the level setting
	NewSample = (INT32)(NoiseValue * 16777216.0f);				// transform to S3.24
	*(pInst->InputMemory) = NewSample;							// copy sine value to memory

	// Store the new sample in the input block only if skipped data is OK
	if (pInst->SkippingCounter != 0)
	{
		// The blocks are not all skipped yet
		pInst->SkippingCounter -= 1;
	}
	else
	{
		pInst->MeasuringMode = 1;
		if (pInst->ProcessingCounter != 0)
		{
			// The processing is not finished yet
			INT32 index = pInst->BlockSize - pInst->ProcessingCounter;
			(pInst->InputBuffer)[index] = NoiseValue;
			pInst->ProcessingCounter -= 1;
		}
		else
		{
			pInst->MeasuringMode = -1;		// This should be the end of simulation
		}
	}

	pInst->OverallCounter += 1;

	return 0;
}

INT32 Tool_FreqResp_WriteSample(void *pInstance)
{
	INT32 NewSample;
	INT32 ReturnValue;
	INT32 index;
	float temp;

	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;

	// Take the new sample
	NewSample = *(pInst->OutputMemory);

	// Take output in account only if the good amount of sample has been skipped
	switch(pInst->MeasuringMode)
	{
	case 0:		// should just skip the data
		ReturnValue = 0;
		break;
	case 1:		// should make measurements now
		temp = (float)NewSample / 16777216.0f;
		index = pInst->BlockSize - pInst->ProcessingCounter - 1;
		(pInst->OutputBuffer)[index] = temp;
		ReturnValue = 0;
		break;
	case -1:	// the simulation should stop now
		ReturnValue = -1;
		break;
	default:	// unsupported case, report bug and stop simulation
		BUG_REPORT("Wrong MeasuringMode value in FreqResp tool write function.\n");
		ReturnValue = -1;
	}

	return ReturnValue;
}





//////////////////////////////////////////////////////////////////////
// file-like function(s)
//////////////////////////////////////////////////////////////////////
INT32 Tool_FreqResp_GetPosition(void *pInstance, INT32 *CurrentPosition)
{
	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;

	*CurrentPosition = pInst->OverallCounter;

	return 0;	//TOOL_NOERROR
}

INT32 Tool_FreqResp_GetFileLength(void *pInstance, INT32 *FileSize)
{
	INT32 index = 1;

	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;

	// This is given by the number of skipped blocks + the measured block
	*FileSize = (pInst->NSkipBlocks + 1) * pInst->BlockSize;
	
	return 0; //TOOL_NOERROR;
}



/*//////////////////////////////////////////////////////////////////////
// control function(s)
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// "text" function(s)
//////////////////////////////////////////////////////////////////////
void FileIO_PCM_GetLastErrorString(CHAR *str, INT32 LastErrorCode, INT32 MaxNumberOfCharacters)
{
	if ((LastErrorCode & FILEIO_BASE_ERR) == FILEIO_BASE_ERR)		// check if the error is a FileIO error
	{
		if ((LastErrorCode & FILEIO_GEN_BASE_ERR) == FILEIO_GEN_BASE_ERR)
		{
			FileIO_GetLastErrorString(str, LastErrorCode, MaxNumberOfCharacters);
			return;
		}
		else if ((LastErrorCode & FILEIO_PCM_BASE_ERR) == FILEIO_PCM_BASE_ERR)
		{
			// do nothing here, wait for the switch below
		}
		else
		{
			strncpy(str, "Error type not known", MaxNumberOfCharacters - 1);
		}
	}
	else															// this error is not a FileIO error
	{
		strncpy(str, "Not a FileIO error", MaxNumberOfCharacters - 1);
		return;
	}


	switch(LastErrorCode)
	{
		case FILEIO_PCM_NOERROR:
			strncpy(str, "PCM: No error occurred", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "PCM: Error not known", MaxNumberOfCharacters - 1);
			break;
	}
	return;
}



void FileIO_PCM_GetHelp(CHAR *str, INT32 LineNumber, INT32 MaxNumberOfCharacters)
{
	switch(LineNumber)
	{
		case 0:
			strncpy(str, "PCM files are read by these functions.", MaxNumberOfCharacters - 1);
			break;
		case 1:
			strncpy(str, "PCM files are raw data encoded in Little Endian format i.e. LSB,MSB", MaxNumberOfCharacters - 1);
			break;
		case 2:
			strncpy(str, "There's no support for 8 bits files and for more than 4GB files.", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "PCM file functions", MaxNumberOfCharacters - 1);
	}
}
*/





//////////////////////////////////////////////////////////////////////
// Helper function(s)
//////////////////////////////////////////////////////////////////////
INT32 Tool_FreqResp_GetResultSize(Tools_FreqResp_Parameters *Params, INT32 *ResultSize)
{
	INT32 index = 1;
	double bs;
	double log2bs;
	double pow2log2bs;
	int BlockSize;


	// Test validity of parameters
	bs = (double)Params->BlockSize;
	log2bs = log(bs) / log(2);
	log2bs = ceil(log2bs);
	pow2log2bs = pow(2, log2bs);
	BlockSize = (int)pow2log2bs;

	if (BlockSize != Params->BlockSize)
	{
		WARNING_1("Tool FreqResp Warning: BlockSize was not a power of two: %i\n", Params->BlockSize);
		WARNING_1("Tool FreqResp Warning: BlockSize has been changed to %i\n", BlockSize);
		Params->BlockSize = BlockSize;
	}

	*ResultSize = (Params->BlockSize / 2);

	return 0;	// NO_ERROR
}

// This function is meant for using the Audacity FFT
/*INT32 Tool_FreqResp_ComputeResults(void *pInstance)
{
	//FFTReal *FFT;
	float *InF_R;
	float *InF_C;
	float *OutF_R;
	float *OutF_C;
	INT32 index;
	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;
	INT32 ResultSize = pInst->BlockSize / 2;
	float MagnIn, MagnOut, PhaseIn, PhaseOut;
	
	// Allocate the FFT with the good size
//	FFT = new FFTReal(pInst->BlockSize);
//	if (FFT == NULL)
//	{
//		ERROR("Unable to allocate memory for the FFT operation.\n");
//		return -1;
//	}

	// Allocate memory for the input buffer result
	InF_R = new float[pInst->BlockSize];
	if (InF_R == NULL)
	{
		ERROR("Unable to allocate memory for the input buffer FFT operation.\n");
//		delete FFT;
		return -1;
	}

	InF_C = new float[pInst->BlockSize];
	if (InF_C == NULL)
	{
		ERROR("Unable to allocate memory for the input buffer FFT operation.\n");
//		delete FFT;
		delete InF_R;
		return -1;
	}

	// Allocate memory for the output buffer result
	OutF_R = new float[pInst->BlockSize];
	if (OutF_R == NULL)
	{
		ERROR("Unable to allocate memory for the output buffer FFT operation.\n");
		delete InF_R;
		delete InF_C;
//		delete FFT;
		return -1;
	}
	OutF_C = new float[pInst->BlockSize];
	if (OutF_C == NULL)
	{
		ERROR("Unable to allocate memory for the output buffer FFT operation.\n");
		delete OutF_R;
		delete InF_R;
		delete InF_C;
//		delete FFT;
		return -1;
	}

	// Apply a window to the signal
//	ApplyHanningWindow(InF, InF, pInst->BlockSize);
//	ApplyHanningWindow(OutF, OutF, pInst->BlockSize);
	WindowFunc(3, pInst->BlockSize, pInst->InputBuffer);
	WindowFunc(3, pInst->BlockSize, pInst->OutputBuffer);

	// Compute the FFTs	
	//FFT->do_fft(InF, pInst->InputBuffer);
	//FFT->do_fft(OutF, pInst->OutputBuffer);
	RealFFT(pInst->BlockSize, pInst->InputBuffer, InF_R, InF_C);
	RealFFT(pInst->BlockSize, pInst->OutputBuffer, OutF_R, OutF_C);

	// Make "computations" for index = 0
	pInst->Results[0].Frequency = 0.0f;
	pInst->Results[0].Amplitude = 20.0f * (float)log10(OutF_R[0] / InF_R[0]);
	pInst->Results[0].Phase = 0.0f;

	// Make "computations" for index = ResultSize
	pInst->Results[ResultSize].Frequency = (float)(ResultSize) * pInst->SamplingFrequency / (float)(pInst->BlockSize);;
	pInst->Results[ResultSize].Amplitude = 20.0f * (float)log10(OutF_R[ResultSize] / InF_R[ResultSize]);
	pInst->Results[ResultSize].Phase = 0.0f;

	// Calculate for each frequency band the amplitude, phase, frequency and store in result vector
	for (index = 1; index < ResultSize; index++)
	{
		// Calculate the frequency value for the given index
		pInst->Results[index].Frequency = (float)index * pInst->SamplingFrequency / (float)(pInst->BlockSize);
		// Calculate the magnitude of the input and output
		double MagnInD  = sqrt( InF_R[index] *  InF_R[index] +  InF_C[index] *  InF_C[index]);
		double MagnOutD = sqrt(OutF_R[index] * OutF_R[index] + OutF_C[index] * OutF_C[index]);
		pInst->Results[index].Amplitude = (float)(20.0 * log10(MagnOutD / MagnInD));
		// Calculate the phase of the input and output
		double PhaseInD  = atan( InF_C[index] /  InF_R[index]);
		double PhaseOutD = atan(OutF_C[index] / OutF_R[index]);
		pInst->Results[index].Phase = (float)(PhaseOutD - PhaseInD);
	}

	PowerSpectrum(pInst->BlockSize, pInst->InputBuffer, InF_R);
	PowerSpectrum(pInst->BlockSize, pInst->OutputBuffer, OutF_R);
	// Calculate for each frequency band the amplitude, phase, frequency and store in result vector
	for (index = 0; index <= ResultSize; index++)
	{
		pInst->Results[index].Amplitude = (float)(20.0 * log10(OutF_R[index] / InF_R[index]));
	}


	// Free the FFT and buffers
	//delete FFT;
	delete InF_R;
	delete InF_C;
	delete OutF_R;
	delete OutF_C;
	
	return 0;
}*/

INT32 Tool_FreqResp_ComputeResults(void *pInstance)
{
	FFTReal *FFT;
	float *InF;
	float *OutF;
	INT32 index;
	Tool_FreqResp_Struct *pInst = (Tool_FreqResp_Struct *)pInstance;
	INT32 ResultSize = pInst->BlockSize / 2;
	double MagnIn, MagnOut, PhaseIn, PhaseOut;
	
	// Allocate the FFT with the good size
	FFT = new FFTReal(pInst->BlockSize);
	if (FFT == NULL)
	{
		ERROR("Unable to allocate memory for the FFT operation.\n");
		return -1;
	}

	// Allocate memory for the input buffer result
	InF = new float[pInst->BlockSize];
	if (InF == NULL)
	{
		ERROR("Unable to allocate memory for the input buffer FFT operation.\n");
		delete FFT;
		return -1;
	}

	// Allocate memory for the output buffer result
	OutF = new float[pInst->BlockSize];
	if (OutF == NULL)
	{
		ERROR("Unable to allocate memory for the output buffer FFT operation.\n");
		delete InF;
		delete FFT;
		return -1;
	}

	// Apply a window to the signal
	ApplyHanningWindow(pInst->InputBuffer, pInst->InputBuffer, pInst->BlockSize);
	ApplyHanningWindow(pInst->OutputBuffer, pInst->OutputBuffer, pInst->BlockSize);

	// Compute the FFTs	
	FFT->do_fft(InF, pInst->InputBuffer);
	FFT->do_fft(OutF, pInst->OutputBuffer);
	FFT->rescale(pInst->InputBuffer);
	FFT->rescale(pInst->OutputBuffer);

	// Make "computations" for index = 0
	pInst->Results[0].Frequency = 0.0f;
	pInst->Results[0].Amplitude = (float)(20.0 * log10(OutF[0] / InF[0]));
	pInst->Results[0].Phase = 0.0f;

	// Calculate for each frequency band the amplitude, phase, frequency and store in result vector
	for (index = 1; index < ResultSize; index++)
	{
		// Calculate the frequency value for the given index
		pInst->Results[index].Frequency = (float)index * pInst->SamplingFrequency / (float)(pInst->BlockSize);
		// Calculate the magnitude of the input and output
		MagnIn  = sqrt( InF[index] *  InF[index] +  InF[ResultSize + index] *  InF[ResultSize + index]);
		MagnOut = sqrt(OutF[index] * OutF[index] + OutF[ResultSize + index] * OutF[ResultSize + index]);
		pInst->Results[index].Amplitude = (float)(20.0 * log10(MagnOut / MagnIn));
		// Calculate the phase of the input and output
		PhaseIn  = atan( InF[ResultSize + index] /  InF[index]);
		PhaseOut = atan(OutF[ResultSize + index] / OutF[index]);
		pInst->Results[index].Phase = (float)(PhaseOut - PhaseIn);
	}

	// Make "computations" for index = ResultSize
	pInst->Results[ResultSize].Frequency = (float)(ResultSize) * pInst->SamplingFrequency / (float)(pInst->BlockSize);;
	double temp = OutF[ResultSize] / InF[ResultSize];
	if (temp < 0.0) pInst->Results[ResultSize].Amplitude = pInst->Results[ResultSize - 1].Amplitude;
	else pInst->Results[ResultSize].Amplitude = (float)(20.0f * log10(temp));
	pInst->Results[ResultSize].Phase = 0.0f;

	// Free the FFT and buffers
	delete FFT;
	delete InF;
	delete OutF;
	
	return 0;
}

INT32 Tool_FreqResp_WriteAmplitudeResults(Tool_FreqResp_Results *results, INT32 ResultSize, CHAR *AmplitudeFileName)
{
	FILE *fAmplitude;
	INT32 index;

	// Open the amplitude file
	fAmplitude = fopen(AmplitudeFileName, "w+t");
	if (fAmplitude == NULL) return -1;

	for(index = 0; index <= ResultSize; index++)
	{
		fprintf(fAmplitude, "%g\t%g\n", results[index].Frequency, results[index].Amplitude);
	}

	fclose(fAmplitude);

	return 0;// NO_ERROR;
}

INT32 Tool_FreqResp_WritePhaseResults(Tool_FreqResp_Results *results, INT32 ResultSize, CHAR *PhaseFileName)
{
	FILE *fPhase;
	INT32 index;

	// Open the phase file
	fPhase = fopen(PhaseFileName, "w+t");
	if (fPhase == NULL) return -1;

	for(index = 0; index <= ResultSize; index++)
	{
		fprintf(fPhase, "%g\t%g\n", results[index].Frequency, results[index].Phase);
	}

	fclose(fPhase);

	return 0;// NO_ERROR;
}
